# vi: ts=4 sw=4 sts=4 expandtab
'''
:mod:`ophyd.utils.epics_pvs` - EPICS-related utilities
======================================================

.. module:: ophyd.utils.epics_pvs
   :synopsis:
'''
from enum import IntEnum
import time as ttime
import ctypes
import threading
import queue
import logging
import warnings
import functools
import numpy as np

import epics

from .errors import DisconnectedError, OpException

__all__ = ['split_record_field',
           'strip_field',
           'record_field',
           'MonitorDispatcher',
           'get_pv_form',
           'set_and_wait',
           'AlarmStatus',
           'AlarmSeverity',
           ]

logger = logging.getLogger(__name__)


class BadPVName(ValueError, OpException):
    ...


class AlarmSeverity(IntEnum):
    NO_ALARM = 0
    MINOR = 1
    MAJOR = 2
    INVALID = 3


class AlarmStatus(IntEnum):
    NO_ALARM = 0
    READ = 1
    WRITE = 2
    HIHI = 3
    HIGH = 4
    LOLO = 5
    LOW = 6
    STATE = 7
    COS = 8
    COMM = 9
    TIMEOUT = 10
    HWLIMIT = 11
    CALC = 12
    SCAN = 13
    LINK = 14
    SOFT = 15
    BAD_SUB = 16
    UDF = 17
    DISABLE = 18
    SIMM = 19
    READ_ACCESS = 20
    WRITE_ACCESS = 21


def validate_pv_name(pv):
    '''Validates that there is not more than 1 '.' in pv

    Parameters
    ----------
    pv : str
        The pv to check

    Raises
    ------
    BadPVName
    '''
    if pv.count('.') > 1:
        raise BadPVName(pv)


def split_record_field(pv):
    '''Splits a pv into (record, field)

    Parameters
    ----------
    pv : str
        the pv to split

    Returns
    -------
    record : str
    field : str
    '''
    if '.' in pv:
        record, field = pv.rsplit('.', 1)
    else:
        record, field = pv, ''

    return record, field


def strip_field(pv):
    '''Strip off the field from a record'''
    return split_record_field(pv)[0]


def record_field(record, field):
    '''Given a record and a field, combine them into
    a pv of the form: record.FIELD
    '''
    record = strip_field(record)
    return '%s.%s' % (record, field.upper())


class MonitorDispatcher(epics.ca.CAThread):
    '''A monitor dispatcher which works with pyepics

    The monitor dispatcher works around having callbacks from libca threads.
    Using epics CA calls (caget, caput, etc.) from those callbacks is not
    possible without this dispatcher workaround.

    ... note:: Without `all_contexts` set, only the callbacks that are run with
        the same context as the the main thread are affected.

    ... note:: Ensure that you call epics.ca.use_initial_context() at startup in
        the main thread

    Parameters
    ----------
    all_contexts : bool, optional
        re-route _all_ callbacks from _any_ context to the dispatcher callback
        thread
    timeout : float, optional
    callback_logger : logging.Logger, optional
        A logger to notify about failed callbacks

    Attributes
    ----------
    main_context : ctypes long
        The main CA context
    callback_logger : logging.Logger
        A logger to notify about failed callbacks
    queue : Queue
        The event queue
    '''

    def __init__(self, all_contexts=False, timeout=0.1,
                 callback_logger=None):
        epics.ca.CAThread.__init__(self, name='monitor_dispatcher')

        self.daemon = True
        self.queue = queue.Queue()

        # The dispatcher thread will stop if this event is set
        self._stop_event = threading.Event()
        self.main_context = epics.ca.current_context()
        self.callback_logger = callback_logger

        self._all_contexts = bool(all_contexts)
        self._timeout = timeout

        self.start()

    def run(self):
        '''The dispatcher itself'''
        self._setup_pyepics(True)

        while not self._stop_event.is_set():
            try:
                callback, args, kwargs = self.queue.get(True, self._timeout)
            except queue.Empty:
                pass
            else:
                try:
                    callback(*args, **kwargs)
                except Exception as ex:
                    if self.callback_logger is not None:
                        self.callback_logger.error(ex, exc_info=ex)

        self._setup_pyepics(False)
        epics.ca.detach_context()

    def stop(self):
        '''Stop the dispatcher thread and re-enable normal callbacks'''
        self._stop_event.set()

    def _setup_pyepics(self, enable):
        # Re-route monitor events to our new handler
        if enable:
            fcn = self._monitor_event
        else:
            fcn = epics.ca._onMonitorEvent

        epics.ca._CB_EVENT = ctypes.CFUNCTYPE(None, epics.dbr.event_handler_args)(fcn)

    def _monitor_event(self, args):
        if self._all_contexts or self.main_context == epics.ca.current_context():
            if callable(args.usr):
                if not hasattr(args.usr, '_disp_tag') or args.usr._disp_tag is not self:
                    args.usr = lambda orig_cb=args.usr, **kwargs: \
                        self.queue.put((orig_cb, [], kwargs))
                    args.usr._disp_tag = self

        return epics.ca._onMonitorEvent(args)


def waveform_to_string(value, type_=str, delim=''):
    '''Convert a waveform that represents a string into an actual Python string

    Parameters
    ----------
    value
        The value to convert
    type_ : type, optional
        Python type to convert to
    delim : str, optional
        delimiter to use when joining string
    '''
    try:
        value = delim.join(chr(c) for c in value)
    except TypeError:
        value = type_(value)

    try:
        value = value[:value.index('\0')]
    except (IndexError, ValueError):
        pass

    return value


def get_pv_form():
    '''Get the PV form that should be used for pyepics

    Due to a bug in certain versions of PyEpics, form='time' cannot be used
    with some large arrays.

    native: gives time.time() timestamps from this machine
    time: gives timestamps from the PVs themselves

    Returns
    -------
    {'native', 'time'}
    '''
    def _fix_git_versioning(in_str):
        return in_str.replace('-g', '+g')

    def _naive_parse_version(version):
        try:
            version = version.lower()

            # Strip off the release-candidate version number (best-effort)
            if 'rc' in version:
                version = version[:version.index('rc')]

            version_tuple = tuple(int(v) for v in version.split('.'))
        except:
            return None

        return version_tuple

    try:
        from pkg_resources import parse_version
    except ImportError:
        parse_version = _naive_parse_version

    version = parse_version(_fix_git_versioning(epics.__version__))

    if version is None:
        warnings.warn('Unrecognized PyEpics version; using local timestamps',
                      ImportWarning)
        return 'native'

    elif version <= parse_version('3.2.3'):
        warnings.warn('PyEpics versions <= 3.2.3 will use local timestamps (version: %s)' %
                      epics.__version__,
                      ImportWarning)
        return 'native'
    else:
        return 'time'


pv_form = get_pv_form()


def records_from_db(fn):
    '''Naively parse db/template files looking for record names

    Returns
    -------
    records : list
        [(record type, record name), ...]
    '''

    ret = []
    for line in open(fn, 'rt').readlines():
        line = line.strip()

        if line.startswith('#'):
            continue

        if not (line.startswith('record') or line.startswith('grecord')):
            continue

        if '(' not in line:
            continue

        line = line[line.index('(') + 1:]
        if ',' not in line:
            continue

        rtype, record = line.split(',', 1)
        rtype = rtype.strip()
        record = record.strip()

        if record.startswith('"'):
            # Surrounded by quotes, easy to parse
            record = record[1:]
            record = record[:record.index('"')]
        else:
            # No quotes, and macros may contain parentheses
            # Find the first non-matching parenthesis and
            # that should denote the end of the record name
            #
            # $(P)$(R)Record)
            #               ^

            in_paren = 0
            for i, c in enumerate(record):
                if c == '(':
                    in_paren += 1
                elif c == ')':
                    in_paren -= 1

                    if in_paren < 0:
                        record = record[:i]
                        break

        ret.append((rtype, record))

    return ret


def raise_if_disconnected(fcn):
    '''Decorator to catch attempted access to disconnected EPICS channels.'''
    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        if self.connected:
            return fcn(self, *args, **kwargs)
        else:
            raise DisconnectedError('{} is not connected'.format(self.name))
    return wrapper


def set_and_wait(signal, val, poll_time=0.01, timeout=10, rtol=None,
                 atol=None):
    """Set a signal to a value and wait until it reads correctly.

    For floating point values, it is strongly recommended to set a tolerance.
    If tolerances are unset, the values will be compared exactly.

    Parameters
    ----------
    signal : EpicsSignal (or any object with `get` and `put`)
    val : object
        value to set signal to
    poll_time : float, optional
        how soon to check whether the value has been successfully set
    timeout : float, optional
        maximum time to wait for value to be successfully set
    rtol : float, optional
        allowed absolute tolerance between the readback and setpoint values
    atol : float, optional
        allowed relative tolerance between the readback and setpoint values

    Raises
    ------
    TimeoutError if timeout is exceeded
    """
    if atol is None and hasattr(signal, 'tolerance'):
        atol = signal.tolerance
    if rtol is None and hasattr(signal, 'rtolerance'):
        rtol = signal.rtolerance

    signal.put(val)
    expiration_time = ttime.time() + timeout if timeout is not None else None
    current_value = signal.get()
    try:
        es = signal.enum_strs
    except AttributeError:
        es = ()

    if atol is not None:
        within_str = ['within {!r}'.format(atol)]
    else:
        within_str = []

    if rtol is not None:
        within_str.append('(relative tolerance of {!r})'.format(rtol))

    if within_str:
        within_str = ' '.join([''] + within_str)
    else:
        within_str = ''

    while not _compare_maybe_enum(val, current_value, es, atol, rtol):
        logger.info("Waiting for %s to be set from %r to %r%s...",
                    signal.name, current_value, val, within_str)
        ttime.sleep(poll_time)
        poll_time *= 2  # logarithmic back-off
        current_value = signal.get()
        if expiration_time is not None and ttime.time() > expiration_time:
            raise TimeoutError("Attempted to set %r to value %r and timed "
                               "out after %r seconds. Current value is %r." %
                               (signal, val, timeout, current_value))


def _compare_maybe_enum(a, b, enums, atol, rtol):
    if enums:
        # convert enum values to strings if necessary first:
        if not isinstance(a, str):
            a = enums[a]
        if not isinstance(b, str):
            b = enums[b]
        # then compare the strings
        return a == b

    # if either relative/absolute tolerance is used, use numpy
    # to compare:
    if atol is not None or rtol is not None:
        return np.allclose(a, b,
                           rtol=rtol if rtol is not None else 1e-5,
                           atol=atol if atol is not None else 1e-8,
                           )
    return a == b


_type_map = {'number': (float, ),
             'array': (np.ndarray, ),
             'string': (str, ),
             'integer': (int, ),
             }


def data_type(val):
    '''Determine data-type of val.

    Returns:
    -----------
    str
        One of ('number', 'array', 'string'), else raises ValueError
    '''
    for json_type, py_types in _type_map.items():
        if type(val) in py_types:
            return json_type
    # no legit type found...
    raise ValueError('{} not a valid type (int, float, ndarray, str)'.format(val))


def data_shape(val):
    '''Determine data-shape (dimensions)

    Returns:
    --------
    list
        Empty list if val is number or string, otherwise list(np.ndarray.shape)
    '''
    for json_type, py_types in _type_map.items():
        if type(val) in py_types:
            if json_type is 'array':
                return list(val.shape)
            else:
                return list()
    raise ValueError('Cannot determine shape of {}'.format(val))




import logging
import unittest

import epics

from ophyd.utils.startup import setup as setup_ophyd


logger = logging.getLogger('ophyd_session_test')
setup_ophyd()


def setup_package():
    pass

def teardown_package():
    pass


def main(is_main):
    # fmt = '%(asctime)-15s [%(levelname)s] %(message)s'
    # logging.basicConfig(format=fmt, level=logging.DEBUG)
    epics.ca.use_initial_context()

    logger = logging.getLogger('logger')
    logger.setLevel(logging.INFO)

    # if you're running a single test as main, chances are you want debug
    # logging:
    logging.getLogger('__main__').setLevel(logging.DEBUG)
    logging.basicConfig()

    if is_main:
        setup_package()
        unittest.main()
        teardown_package()



import logging
import pytest

from ophyd import QuadEM, Component as Cpt, Signal
from ophyd.areadetector.plugins import ImagePlugin, StatsPlugin
from .test_signal import using_fake_epics_pv


logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
@using_fake_epics_pv
def quadem():
    class FakeStats(StatsPlugin):
        plugin_type = Cpt(Signal, value=StatsPlugin._plugin_type)
        nd_array_port = Cpt(Signal, value='NSLS_EM')

    class FakeImage(ImagePlugin):
        plugin_type = Cpt(Signal, value=ImagePlugin._plugin_type)
        nd_array_port = Cpt(Signal, value='NSLS_EM')

    class FakeQuadEM(QuadEM):
        image = Cpt(FakeImage, 'image1:')
        current1 = Cpt(FakeStats, 'Current1:')
        current2 = Cpt(FakeStats, 'Current2:')
        current3 = Cpt(FakeStats, 'Current3:')
        current4 = Cpt(FakeStats, 'Current4:')

    em = FakeQuadEM('quadem:', name='quadem')

    ''' Beware: Ugly Hack below

        tl;dr
        Set Signal._read_pv = Signal._write_pv in order for
        set_and_wait() to succeed in calls from Device.stage()

        The raison d'etre here is a limitation of FakeEpicsPV,
        or rather a limitation of the test harness:
            Since the QuadEM is based on areadetector, it uses
            EpicsSignalWithRBV in several places. The test harness
            monkey-patches epics.PV with FakeEpicsPV, which means
            that get/put are routed to two different pvs, which in turn,
            means that set_and_wait() will never be successful for
            EpicsSignalWithRBVs... :-(
    '''
    for sig in em.stage_sigs:
        sig = getattr(em, sig)
        sig._read_pv = sig._write_pv

    for sig in em.image.stage_sigs:
        sig = getattr(em.image, sig)
        sig._read_pv = sig._write_pv
    em.image.enable._read_pv = em.image.enable._write_pv

    for sig in em.current1.stage_sigs:
        sig = getattr(em.current1, sig)
        sig._read_pv = sig._write_pv
    em.current1.enable._read_pv = em.current1.enable._write_pv

    for sig in em.current2.stage_sigs:
        sig = getattr(em.current2, sig)
        sig._read_pv = sig._write_pv
    em.current2.enable._read_pv = em.current2.enable._write_pv

    for sig in em.current3.stage_sigs:
        sig = getattr(em.current3, sig)
        sig._read_pv = sig._write_pv
    em.current3.enable._read_pv = em.current3.enable._write_pv

    for sig in em.current4.stage_sigs:
        sig = getattr(em.current4, sig)
        sig._read_pv = sig._write_pv
    em.current4.enable._read_pv = em.current4.enable._write_pv
    ''' End: Ugly Hack '''

    for sig in ['image'] + ['current{}'.format(j) for j in range(1, 5)]:
        cpt = getattr(em, sig)
        cpt.port_name._read_pv.put(sig.upper())

    em.wait_for_connection()

    return em


def test_connected(quadem):
    assert quadem.connected


@using_fake_epics_pv
def test_scan_point(quadem):
    assert quadem._staged.value == 'no'

    quadem.stage()
    assert quadem._staged.value == 'yes'

    quadem.trigger()
    quadem.unstage()
    assert quadem._staged.value == 'no'


@using_fake_epics_pv
def test_reading(quadem):
    assert 'current1.mean_value' in quadem.read_attrs

    desc = quadem.describe()
    desc_keys = list(desc['quadem_current1_mean_value'].keys())
    assert (set(desc_keys) == set(['dtype', 'precision', 'shape', 'source',
                                   'units', 'lower_ctrl_limit',
                                   'upper_ctrl_limit']))

    vals = quadem.read()
    assert 'quadem_current1_mean_value' in vals
    assert (set(('value', 'timestamp')) ==
            set(vals['quadem_current1_mean_value'].keys()))




import logging

from collections import OrderedDict

from .signal import (EpicsSignal, EpicsSignalRO)
from .device import Device
from .device import (Component as C, DynamicDeviceComponent as DDC)

logger = logging.getLogger(__name__)


def _scaler_fields(attr_base, field_base, range_, **kwargs):
    defn = OrderedDict()
    for i in range_:
        attr = '{attr}{i}'.format(attr=attr_base, i=i)
        suffix = '{field}{i}'.format(field=field_base, i=i)
        defn[attr] = (EpicsSignalRO, suffix, kwargs)

    return defn


class EpicsScaler(Device):
    '''SynApps Scaler Record interface'''

    count = C(EpicsSignal, '.CNT', trigger_value=1)
    count_mode = C(EpicsSignal, '.CONT', string=True)
    time = C(EpicsSignal, '.T')
    preset_time = C(EpicsSignal, '.TP')
    auto_count_time = C(EpicsSignal, '.TP1')
    channels = DDC(_scaler_fields('chan', '.S', range(1, 33)))
    presets = DDC(_scaler_fields('preset', '.PR', range(1, 33)))
    gates = DDC(_scaler_fields('gate', '.G', range(1, 33)))

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 name=None, parent=None, **kwargs):
        if read_attrs is None:
            read_attrs = ['channels', 'time']

        if configuration_attrs is None:
            configuration_attrs = ['preset_time', 'presets', 'gates']

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs,
                         name=name, parent=parent, **kwargs)

        self.stage_sigs.update([('count_mode', 0)])




import logging
import unittest
# from unittest.mock import Mock
from contextlib import contextmanager

import ophyd.commands
from ophyd import EpicsMotor
from ophyd.commands import (mov, movr, set_pos, wh_pos, set_lm)
from ophyd.commands import (log_pos, log_pos_diff, log_pos_mov,
                            get_all_positioners, get_logbook, setup_ophyd)
from .config import motor_recs

logger = logging.getLogger(__name__)
mtr = None


def setUpModule():
    global mtr

    setup_ophyd()

    mtr = EpicsMotor(motor_recs[0])
    mtr.wait_for_connection()


def tearDownModule():
    pass


class SimpleOlogClient:
    def log(self, text=None, logbooks=None, tags=None, properties=None,
            attachments=None, verify=True, ensure=False):
        return None

    def find(self, id=None, **kwargs):
        if id != 1:
            return []

        pos = {'objects': ("[EpicsMotor('{0}', name='mtr')]"
                           "".format(motor_recs[0])),
               'values': 'dict(mtr=1.0)',
               }

        entry = dict(properties={'OphydPositioners': pos})
        return [entry]


def _get_logbook():
    return SimpleOlogClient()
    # import pyOlog
    # return Mock(pyOlog.SimpleOlogClient, instance=True)


@contextmanager
def mock_logbook():
    get_logbook = ophyd.commands.get_logbook
    ophyd.commands.get_logbook = _get_logbook
    try:
        yield
    finally:
        ophyd.commands.get_logbook = get_logbook


class Commands(unittest.TestCase):
    def test_move_absolute(self):
        global mtr

        self.assertRaises(TypeError, mov, 'not_a_positioner', 0.0)
        mov(mtr, 0.0)

    def test_move_relative(self):
        global mtr

        self.assertRaises(TypeError, movr, 'not_a_positioner', 0.0)
        mov(mtr, 0.0)

    def test_wh_pos(self):
        global mtr
        self.assertRaises(TypeError, wh_pos, 'not_a_positioner')

        wh_pos()
        wh_pos(mtr)
        wh_pos([mtr] * 10)
        self.assertRaises(TypeError, wh_pos, [mtr, None])

    def test_set_limits(self):
        global mtr
        set_lm(mtr, mtr.limits)
        self.assertRaises(TypeError, set_lm, mtr, None)
        set_lm([mtr, mtr], [mtr.limits, mtr.limits])
        self.assertRaises(TypeError, set_lm, [mtr, mtr], [mtr.limits, None])

    def test_set_position(self):
        global mtr
        set_pos(mtr, mtr.position)
        self.assertRaises(TypeError, set_pos, mtr, None)
        set_pos([mtr, mtr], [mtr.position, mtr.position])

    def test_get_all_positioners(self):
        # TODO: mock up IPython session
        get_all_positioners()

    def test_get_logbook(self):
        # TODO: can't test without IPython session
        get_logbook()
        with mock_logbook():
            get_logbook()

    def test_log_pos(self):
        global mtr
        log_pos()

        with mock_logbook():
            log_pos()
            log_pos(mtr)
            log_pos([mtr, mtr])
            self.assertRaises(TypeError, log_pos, [mtr, None])

    def test_log_pos_diff(self):
        global mtr
        with mock_logbook():
            log_pos_diff(id=1, positioners=[mtr])
            self.assertRaises(ValueError, log_pos_diff, id=2, positioners=[mtr])


from . import main
is_main = (__name__ == '__main__')
main(is_main)



import time
import logging
import unittest
import pytest

from copy import copy

import epics
from ophyd import (PseudoPositioner, PseudoSingle, EpicsMotor)
from ophyd import (Component as C)
from ophyd.utils import ExceptionBundle


logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger('ophyd.pseudopos').setLevel(logging.DEBUG)


def tearDownModule():
    if __name__ == '__main__':
        epics.ca.destroy_context()

    logger.debug('Cleaning up')
    logging.getLogger('ophyd.pseudopos').setLevel(logging.INFO)


motor_recs = ['XF:31IDA-OP{Tbl-Ax:X1}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X2}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X3}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X4}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X5}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X6}Mtr',
              ]


class Pseudo3x3(PseudoPositioner):
    pseudo1 = C(PseudoSingle, '', limits=(-10, 10), egu='a')
    pseudo2 = C(PseudoSingle, '', limits=(-10, 10), egu='b')
    pseudo3 = C(PseudoSingle, '', limits=None, egu='c')
    real1 = C(EpicsMotor, motor_recs[0])
    real2 = C(EpicsMotor, motor_recs[1])
    real3 = C(EpicsMotor, motor_recs[2])

    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(real1=-pseudo_pos.pseudo1,
                                 real2=-pseudo_pos.pseudo2,
                                 real3=-pseudo_pos.pseudo3)

    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(pseudo1=real_pos.real1,
                                   pseudo2=real_pos.real2,
                                   pseudo3=real_pos.real3)


class Pseudo1x3(PseudoPositioner):
    pseudo1 = C(PseudoSingle, limits=(-10, 10))
    real1 = C(EpicsMotor, motor_recs[0])
    real2 = C(EpicsMotor, motor_recs[1])
    real3 = C(EpicsMotor, motor_recs[2])

    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(real1=-pseudo_pos.pseudo1,
                                 real2=-pseudo_pos.pseudo1,
                                 real3=-pseudo_pos.pseudo1)

    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(pseudo1=-real_pos.real1)


class FaultyStopperEpicsMotor(EpicsMotor):
    def stop(self, *, success=False):
        raise RuntimeError('Expected exception')


class FaultyPseudo1x3(Pseudo1x3):
    real1 = C(FaultyStopperEpicsMotor, motor_recs[0])


class PseudoPosTests(unittest.TestCase):
    def test_onlypseudo(self):
        # can't instantiate it on its own
        self.assertRaises(TypeError, PseudoPositioner, 'prefix')

    def test_position_wrapper(self):
        pseudo = Pseudo3x3('', name='mypseudo', concurrent=False)

        test_pos = pseudo.PseudoPosition(pseudo1=1, pseudo2=2, pseudo3=3)
        extra_kw = dict(a=3, b=4, c=6)

        # positional arguments
        self.assertEqual(pseudo.to_pseudo_tuple(1, 2, 3, **extra_kw),
                         (test_pos, extra_kw))
        # sequence
        self.assertEqual(pseudo.to_pseudo_tuple((1, 2, 3), **extra_kw),
                         (test_pos, extra_kw))
        # correct type
        self.assertEqual(pseudo.to_pseudo_tuple(test_pos, **extra_kw),
                         (test_pos, extra_kw))
        # kwargs
        self.assertEqual(pseudo.to_pseudo_tuple(pseudo1=1, pseudo2=2,
                                                pseudo3=3, **extra_kw),
                         (test_pos, extra_kw))

        # too many positional arguments
        self.assertRaises(ValueError, pseudo.to_pseudo_tuple, 1, 2, 3, 4)
        # too few positional arguments
        self.assertRaises(ValueError, pseudo.to_pseudo_tuple, 1, 2)
        # too few kwargs
        self.assertRaises(ValueError, pseudo.to_pseudo_tuple, pseudo1=1,
                          pseudo2=2)
        # valid kwargs, but passing in args too
        self.assertRaises(ValueError, pseudo.to_pseudo_tuple, 1, pseudo1=1,
                          pseudo2=2, pseudo3=3)

    def test_multi_sequential(self):
        def done(**kwargs):
            logger.debug('** Finished moving (%s)', kwargs)

        pseudo = Pseudo3x3('', name='mypseudo', concurrent=False)
        pseudo.wait_for_connection()

        self.assertEqual(pseudo.egu, 'a, b, c')

        print('real1', pseudo.real1.prefix)
        print('real2', pseudo.real2.prefix)
        print('real3', pseudo.real3.prefix)

        print('real position is made of', pseudo.RealPosition._fields)
        print('pseudo position is made of', pseudo.PseudoPosition._fields)
        print('current pseudo position', pseudo.position)
        print(repr(pseudo))
        print(str(pseudo))

        pos2 = pseudo.PseudoPosition(pseudo1=0, pseudo2=0, pseudo3=0)
        pseudo.move(pos2, wait=True)
        print('moved to', pseudo.position)
        print('-----------------')
        time.sleep(1.0)
        pos1 = pseudo.PseudoPosition(pseudo1=.1, pseudo2=.2, pseudo3=.3)
        pseudo.move(pos1, wait=True)
        print('moved to', pseudo.position)

        pseudo.real1.move(0, wait=True)
        pseudo.real2.move(0, wait=True)
        pseudo.real3.move(0, wait=True)

        pseudo.pseudo1.stop()

        pseudo.real3.move(0, wait=True)

    def test_faulty_stopper(self):
        pseudo = FaultyPseudo1x3('', name='mypseudo', concurrent=False)
        pseudo.wait_for_connection()

        with pytest.raises(ExceptionBundle):
            # smoke-testing for coverage
            pseudo.pseudo1.stop()

    def test_limits(self):
        pseudo = Pseudo3x3('', name='mypseudo', concurrent=True)
        self.assertEquals(pseudo.limits, ((-10, 10), (-10, 10), (0, 0)))
        self.assertEquals(pseudo.low_limit, (-10, -10, 0))
        self.assertEquals(pseudo.high_limit, (10, 10, 0))

    def test_read_describe(self):
        pseudo = Pseudo3x3('', name='mypseudo', concurrent=True)
        desc_dict = pseudo.describe()
        print(desc_dict)
        desc_keys = ['source', 'upper_ctrl_limit', 'lower_ctrl_limit', 'shape',
                     'dtype', 'units']

        for key in desc_keys:
            self.assertIn(key, desc_dict['mypseudo_pseudo3'])

        read_dict = pseudo.read()
        print(read_dict)
        read_keys = ['value', 'timestamp']
        for key in read_keys:
            self.assertIn(key, read_dict['mypseudo_pseudo3'])

        self.assertEqual(pseudo.read().keys(), pseudo.describe().keys())

    def test_multi_concurrent(self):
        def done(**kwargs):
            logger.debug('** Finished moving (%s)', kwargs)

        pseudo = Pseudo3x3('', name='mypseudo', concurrent=True,
                           settle_time=0.1, timeout=25.0)
        self.assertIs(pseudo.sequential, False)
        self.assertIs(pseudo.concurrent, True)
        self.assertEqual(pseudo.settle_time, 0.1)
        self.assertEqual(pseudo.timeout, 25.0)
        pseudo.wait_for_connection()

        self.assertTrue(pseudo.connected)
        self.assertEqual(tuple(pseudo.pseudo_positioners),
                         (pseudo.pseudo1, pseudo.pseudo2, pseudo.pseudo3))
        self.assertEqual(tuple(pseudo.real_positioners),
                         (pseudo.real1, pseudo.real2, pseudo.real3))

        logger.info('Move to (.2, .2, .2), which is (-.2, -.2, -.2) for real '
                    'motors')
        pseudo.move(pseudo.PseudoPosition(.2, .2, .2), wait=True)
        logger.info('Position is: %s (moving=%s)', pseudo.position,
                    pseudo.moving)

        pseudo.check_value((2, 2, 2))
        pseudo.check_value(pseudo.PseudoPosition(2, 2, 2))
        try:
            pseudo.check_value((2, 2, 2, 3))
        except ValueError as ex:
            logger.info('Check value failed, as expected (%s)', ex)

        real1 = pseudo.real1
        pseudo1 = pseudo.pseudo1

        try:
            pseudo.check_value((real1.high_limit + 1, 2, 2))
        except ValueError as ex:
            logger.info('Check value failed, as expected (%s)', ex)

        ret = pseudo.move((2, 2, 2), wait=False, moved_cb=done)
        self.assertEqual(ret.settle_time, 0.1)
        while not ret.done:
            logger.info('Pos=%s %s (err=%s)', pseudo.position, ret, ret.error)
            time.sleep(0.1)

        logger.info('Single pseudo axis: %s', pseudo1)

        pseudo1.move(0, wait=True)

        self.assertEquals(pseudo1.target, 0)
        pseudo1.sync()
        self.assertEquals(pseudo1.target, pseudo1.position)

        # coverage
        pseudo1._started_moving

        try:
            pseudo1.check_value(real1.high_limit + 1)
        except ValueError as ex:
            logger.info('Check value for single failed, as expected (%s)', ex)

        logger.info('Move pseudo1 to 0, position=%s', pseudo.position)
        logger.info('pseudo1 = %s', pseudo1.position)

        def single_sub(**kwargs):
            # logger.info('Single sub: %s', kwargs)
            pass

        pseudo1.subscribe(single_sub, pseudo1.SUB_READBACK)

        ret = pseudo1.move(1, wait=False)
        self.assertEqual(pseudo.timeout, ret.timeout)
        while not ret.done:
            logger.info('pseudo1.pos=%s Pos=%s %s (err=%s)', pseudo1.position,
                        pseudo.position, ret, ret.error)
            time.sleep(0.1)

        logger.info('pseudo1.pos=%s Pos=%s %s (err=%s)', pseudo1.position,
                    pseudo.position, ret, ret.error)

        copy(pseudo)
        pseudo.read()
        pseudo.describe()
        pseudo.read_configuration()
        pseudo.describe_configuration()

        repr(pseudo)
        str(pseudo)

        pseudo.pseudo1.read()
        pseudo.pseudo1.describe()
        pseudo.pseudo1.read_configuration()
        pseudo.pseudo1.describe_configuration()

    def test_single_pseudo(self):
        def done(**kwargs):
            logger.debug('** Finished moving (%s)', kwargs)

        logger.info('------- Sequential, single pseudo positioner')
        pos = Pseudo1x3('', name='mypseudo', concurrent=False)

        reals = pos._real

        logger.info('Move to .2, which is (-.2, -.2, -.2) for real motors')
        pos.move((.2, ), wait=True)
        logger.info('Position is: %s (moving=%s)', pos.position, pos.moving)
        logger.info('Real positions: %s', [real.position for real in reals])

        logger.info('Move to -.2, which is (.2, .2, .2) for real motors')
        pos.move((-.2, ), wait=True)
        logger.info('Position is: %s (moving=%s)', pos.position, pos.moving)
        logger.info('Real positions: %s', [real.position for real in reals])

        copy(pos)
        pos.read()
        pos.describe()
        repr(pos)
        str(pos)


from . import main
is_main = (__name__ == '__main__')
main(is_main)



import time as ttime
import logging
import textwrap
from enum import Enum
from collections import (OrderedDict, namedtuple)

from .ophydobj import OphydObject
from .status import DeviceStatus
from .utils import (ExceptionBundle, set_and_wait, RedundantStaging)

logger = logging.getLogger(__name__)


class Staged(Enum):
    """Three-state switch"""
    yes = 'yes'
    no = 'no'
    partially = 'partially'


class Component:
    '''A descriptor representing a device component (or signal)

    Unrecognized keyword arguments will be passed directly to the component
    class initializer.

    Parameters
    ----------
    cls : class
        Class of signal to create.  The required signature of
        `cls.__init__` is (if `suffix` is given)::

            def __init__(self, pv_name, parent=None, **kwargs):

        or (if suffix is None) ::

            def __init__(self, parent=None, **kwargs):

        The class may have a `wait_for_connection()` which is called
        during the component instance creation.

    suffix : str, optional
        The PV suffix, which gets appended onto the device prefix to
        generate the final PV that the instance component will bind to.

    lazy : bool, optional
        Lazily instantiate the signal. If False, the signal will be
        instantiated upon component instantiation
    trigger_value : any, optional
        Mark as a signal to be set on trigger. The value is sent to the signal
        at trigger time.
    add_prefix : sequence, optional
        Keys in the kwargs to prefix with the Device PV prefix during
        creation of the component instance.
        Defaults to ('suffix', 'write_pv', )
    doc : str, optional
        string to attach to component DvcClass.component.__doc__
    '''

    def __init__(self, cls, suffix=None, *, lazy=False, trigger_value=None,
                 add_prefix=None, doc=None, **kwargs):
        self.attr = None  # attr is set later by the device when known
        self.cls = cls
        self.kwargs = kwargs
        self.lazy = lazy
        self.suffix = suffix
        self.doc = doc
        self.trigger_value = trigger_value  # TODO discuss

        if add_prefix is None:
            add_prefix = ('suffix', 'write_pv')

        self.add_prefix = tuple(add_prefix)

    def maybe_add_prefix(self, instance, kw, suffix):
        """Add prefix to a suffix if kw is in self.add_prefix

        Parameters
        ----------
        instance : Device
            The instance to extract the prefix to maybe append to the
            suffix from.

        kw : str
            The key of associated with the suffix.  If this key is
            self.add_prefix than prepend the prefix to the suffix and
            return, else just return the suffix.

        suffix : str
            The suffix to maybe have something prepended to.

        Returns
        -------
        str
        """
        if kw in self.add_prefix:
            return '{prefix}{suffix}'.format(prefix=instance.prefix,
                                             suffix=suffix)
        return suffix

    def create_component(self, instance):
        '''Create a component for the instance'''
        kwargs = self.kwargs.copy()
        kwargs['name'] = '{}_{}'.format(instance.name, self.attr)

        for kw, val in list(kwargs.items()):
            kwargs[kw] = self.maybe_add_prefix(instance, kw, val)

        if self.suffix is not None:
            pv_name = self.maybe_add_prefix(instance, 'suffix', self.suffix)
            cpt_inst = self.cls(pv_name, parent=instance, **kwargs)
        else:
            cpt_inst = self.cls(parent=instance, **kwargs)

        if self.lazy and hasattr(self.cls, 'wait_for_connection'):
            cpt_inst.wait_for_connection()

        return cpt_inst

    def make_docstring(self, parent_class):
        if self.doc is not None:
            return self.doc

        doc = ['{} attribute'.format(self.__class__.__name__),
               '::',
               '',
               ]

        doc.append(textwrap.indent(repr(self), prefix=' ' * 4))
        doc.append('')
        return '\n'.join(doc)

    def __repr__(self):
        kw_str = ', '.join('{}={!r}'.format(k, v)
                           for k, v in self.kwargs.items())
        if self.suffix is not None:
            suffix_str = '{!r}'.format(self.suffix)
            if self.kwargs:
                suffix_str += ', '
        else:
            suffix_str = ''

        if suffix_str or kw_str:
            arg_str = ', {}{}'.format(suffix_str, kw_str)
        else:
            arg_str = ''

        return ('{self.__class__.__name__}({self.cls.__name__}{arg_str})'
                ''.format(self=self, arg_str=arg_str))

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.attr not in instance._signals:
            instance._signals[self.attr] = self.create_component(instance)

        return instance._signals[self.attr]

    def __set__(self, instance, owner):
        raise RuntimeError('Use .put()')


class FormattedComponent(Component):
    '''A Component which takes a dynamic format string

    This differs from Component in that the parent prefix is not automatically
    added onto the Component suffix. Additionally, `str.format()` style strings
    are accepted, allowing access to Device instance attributes:

    >>> from ophyd import (Component as C, FormattedComponent as FC)
    >>> class MyDevice(Device):
    ...     # A normal component, where 'suffix' is added to prefix verbatim
    ...     cpt = C(EpicsSignal, 'suffix')
    ...     # A formatted component, where 'self' refers to the Device instance
    ...     ch = FC(EpicsSignal, '{self.prefix}{self._ch_name}')
    ...
    ...     def __init__(self, prefix, ch_name=None, **kwargs):
    ...         self._ch_name = ch_name
    ...         super().__init__(prefix, **kwargs)

    >>> dev = MyDevice('prefix:', ch_name='some_channel', name='dev')
    >>> print(dev.cpt.pvname)
    prefix:suffix
    >>> print(dev.ch.pvname)
    prefix:some_channel

    For additional documentation, refer to Component.
    '''

    def maybe_add_prefix(self, instance, kw, suffix):
        if kw not in self.add_prefix:
            return suffix

        return suffix.format(self=instance)


class DynamicDeviceComponent:
    '''An Device component that dynamically creates a OphyDevice

    Parameters
    ----------
    defn : OrderedDict
        The definition of all attributes to be created, in the form of:
            defn['attribute_name'] = (SignalClass, pv_suffix, keyword_arg_dict)
        This will create an attribute on the sub-device of type `SignalClass`,
        with a suffix of pv_suffix, which looks something like this:
            parent.attribute_name = SignalClass(pv_suffix, **keyword_arg_dict)
        Keep in mind that this is actually done in the metaclass creation, and
        not exactly as written above.
    clsname : str, optional
        The name of the class to be generated
        This defaults to {parent_name}{this_attribute_name.capitalize()}
    doc : str, optional
        The docstring to put on the dynamically generated class
    '''

    def __init__(self, defn, *, clsname=None, doc=None):
        self.defn = defn
        self.clsname = clsname
        self.attr = None  # attr is set later by the device when known
        self.lazy = False
        self.doc = doc

        # TODO: component compatibility
        self.trigger_value = None
        self.attrs = list(defn.keys())

    def make_docstring(self, parent_class):
        if self.doc is not None:
            return self.doc

        doc = ['{} comprised of'.format(self.__class__.__name__),
               '::',
               '',
               ]

        doc.append(textwrap.indent(repr(self), prefix=' ' * 4))
        doc.append('')
        return '\n'.join(doc)

    def __repr__(self):
        doc = []
        for attr, (cls, suffix, kwargs) in self.defn.items():
            kw_str = ', '.join('{}={!r}'.format(k, v)
                               for k, v in kwargs.items())
            if suffix is not None:
                suffix_str = '{!r}'.format(suffix)
                if kwargs:
                    suffix_str += ', '
            else:
                suffix_str = ''

            if suffix_str or kw_str:
                arg_str = ', {}{}'.format(suffix_str, kw_str)
            else:
                arg_str = ''

            doc.append('{attr} = Component({cls.__name__}{arg_str})'
                       ''.format(attr=attr, cls=cls, arg_str=arg_str))

        return '\n'.join(doc)

    def create_attr(self, attr_name):
        cls, suffix, kwargs = self.defn[attr_name]
        inst = Component(cls, suffix, **kwargs)
        inst.attr = attr_name
        return inst

    def create_component(self, instance):
        '''Create a component for the instance'''
        clsname = self.clsname
        if clsname is None:
            # make up a class name based on the instance's class name
            clsname = ''.join((instance.__class__.__name__,
                               self.attr.capitalize()))

            # TODO: and if the attribute has any underscores, convert that to
            #       camelcase

        docstring = self.doc
        if docstring is None:
            docstring = '{} sub-device'.format(clsname)

        clsdict = OrderedDict(__doc__=docstring)

        for attr in self.defn.keys():
            clsdict[attr] = self.create_attr(attr)

        attrs = set(self.defn.keys())
        inst_read = set(instance.read_attrs)
        if self.attr in inst_read:
            # if the sub-device is in the read list, then add all attrs
            read_attrs = attrs
        else:
            # otherwise, only add the attributes that exist in the sub-device
            # to the read_attrs list
            read_attrs = inst_read.intersection(attrs)

        cls = type(clsname, (Device, ), clsdict)
        return cls(instance.prefix, read_attrs=list(read_attrs),
                   name='{}_{}'.format(instance.name, self.attr),
                   parent=instance)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.attr not in instance._signals:
            instance._signals[self.attr] = self.create_component(instance)

        return instance._signals[self.attr]

    def __set__(self, instance, owner):
        raise RuntimeError('Use .put()')


class ComponentMeta(type):
    '''Creates attributes for Components by inspecting class definition'''

    @classmethod
    def __prepare__(self, name, bases):
        '''Prepare allows the class attribute dictionary to be ordered as
        defined by the user'''
        return OrderedDict()

    def __new__(cls, name, bases, clsdict):
        clsobj = super().__new__(cls, name, bases, clsdict)

        # This attrs are defined at instanitation time and must not
        # collide with class attributes.
        INSTANCE_ATTRS = ['name', 'parent', 'signal_names', '_signals',
                          'read_attrs', 'configuration_attrs', '_sig_attrs',
                          '_sub_devices']
        # These attributes are part of the bluesky interface and cannot be
        # used as component names.
        RESERVED_ATTRS = ['read', 'describe', 'trigger',
                          'configure', 'read_configuration',
                          'describe_configuration', 'describe_collect',
                          'set', 'stage', 'unstage', 'pause', 'resume',
                          'kickoff', 'complete', 'collect', 'position', 'stop',
                          # from OphydObject
                          'subscribe', 'clear_sub', 'event_types', 'root']
        for attr in INSTANCE_ATTRS:
            if attr in clsdict:
                raise TypeError("The attribute name %r is reserved for "
                                "use by the Device class. Choose a different "
                                "name." % attr)

        clsobj._sig_attrs = OrderedDict()
        for base in reversed(bases):
            if not hasattr(base, '_sig_attrs'):
                continue

            for attr, cpt in base._sig_attrs.items():
                clsobj._sig_attrs[attr] = cpt

        # map component classes to their attribute names from this class
        for attr, value in clsdict.items():
            if isinstance(value, (Component, DynamicDeviceComponent)):
                if attr in RESERVED_ATTRS:
                    raise TypeError("The attribute name %r is part of the "
                                    "bluesky interface and cannot be used as "
                                    "the name of a component. Choose a "
                                    "different name." % attr)
                clsobj._sig_attrs[attr] = value

        for cpt_attr, cpt in clsobj._sig_attrs.items():
            # Notify the component of their attribute name
            cpt.attr = cpt_attr

        # List Signal attribute names.
        clsobj.signal_names = list(clsobj._sig_attrs.keys())

        # The namedtuple associated with the device
        clsobj._device_tuple = namedtuple(name + 'Tuple', clsobj.signal_names,
                                          rename=True)

        # Finally, create all the component docstrings
        for cpt in clsobj._sig_attrs.values():
            cpt.__doc__ = cpt.make_docstring(clsobj)

        # List the attributes that are Devices (not Signals).
        # This list is used by stage/unstage. Only Devices need to be staged.
        clsobj._sub_devices = []
        for attr, cpt in clsobj._sig_attrs.items():
            if isinstance(cpt, Component) and not issubclass(cpt.cls, Device):
                continue
            clsobj._sub_devices.append(attr)

        return clsobj


# These stub 'Interface' classes are the apex of the mro heirarchy for
# their respective methods. They make multiple interitance more
# forgiving, and let us define classes that customize these methods
# but are not full Devices.


class BlueskyInterface:
    """Classes that inherit from this can safely customize the
    these methods without breaking mro."""
    def __init__(self, *args, **kwargs):
        # Subclasses can populate this with (signal, value) pairs, to be
        # set by stage() and restored back by unstage().
        self.stage_sigs = OrderedDict()

        self._staged = Staged.no
        self._original_vals = OrderedDict()
        super().__init__(*args, **kwargs)

    def trigger(self):
        pass

    def read(self):
        return OrderedDict()

    def describe(self):
        return OrderedDict()

    def stage(self):
        """
        Prepare the device to be triggered.

        Returns
        -------
        devices : list
            list including self and all child devices staged
        """
        if self._staged == Staged.no:
            pass  # to short-circuit checking individual cases
        elif self._staged == Staged.yes:
            raise RedundantStaging("Device {!r} is already staged. "
                                   "Unstage it first.".format(self))
        elif self._staged == Staged.partially:
            raise RedundantStaging("Device {!r} has been partially staged. "
                                   "Maybe the most recent unstaging "
                                   "encountered an error before finishing. "
                                   "Try unstaging again.".format(self))
        logger.debug("Staging %s", self.name)
        self._staged = Staged.partially

        # Resolve any stage_sigs keys given as strings: 'a.b' -> self.a.b
        stage_sigs = OrderedDict()
        for k, v in self.stage_sigs.items():
            if isinstance(k, str):
                # Device.__getattr__ handles nested attr lookup
                stage_sigs[getattr(self, k)] = v
            else:
                stage_sigs[k] = v

        # Read current values, to be restored by unstage()
        original_vals = {sig: sig.get() for sig, _ in stage_sigs.items()}

        # We will add signals and values from original_vals to
        # self._original_vals one at a time so that
        # we can undo our partial work in the event of an error.

        # Apply settings.
        devices_staged = []
        try:
            for sig, val in stage_sigs.items():
                logger.debug("Setting %s to %r (original value: %r)", self.name,
                             val, original_vals[sig])
                set_and_wait(sig, val)
                # It worked -- now add it to this list of sigs to unstage.
                self._original_vals[sig] = original_vals[sig]
            devices_staged.append(self)

            # Call stage() on child devices.
            for attr in self._sub_devices:
                device = getattr(self, attr)
                if hasattr(device, 'stage'):
                    device.stage()
                    devices_staged.append(device)
        except Exception:
            logger.debug("An exception was raised while staging %s or "
                         "one of its children. Attempting to restore "
                         "original settings before re-raising the "
                         "exception.", self.name)
            self.unstage()
            raise
        else:
            self._staged = Staged.yes
        return devices_staged

    def unstage(self):
        """
        Restore the device to 'standby'.

        Multiple calls (without a new call to 'stage') have no effect.

        Returns
        -------
        devices : list
            list including self and all child devices unstaged
        """
        logger.debug("Unstaging %s", self.name)
        self._staged = Staged.partially
        devices_unstaged = []

        # Call unstage() on child devices.
        for attr in self._sub_devices[::-1]:
            device = getattr(self, attr)
            if hasattr(device, 'unstage'):
                device.unstage()
                devices_unstaged.append(device)

        # Restore original values.
        for sig, val in reversed(list(self._original_vals.items())):
            logger.debug("Setting %s back to its original value: %r)", self.name,
                         val)
            set_and_wait(sig, val)
            self._original_vals.pop(sig)
        devices_unstaged.append(self)

        self._staged = Staged.no
        return devices_unstaged

    def pause(self):
        pass

    def resume(self):
        pass


class GenerateDatumInterface:
    """Classes that inherit from this can safely customize the
    `generate_datum` method without breaking mro. If used along with the
    BlueskyInterface, inherit from this second."""
    def generate_datum(self, key, timestamp):
        pass


class Device(BlueskyInterface, OphydObject, metaclass=ComponentMeta):
    """Base class for device objects

    This class provides attribute access to one or more Signals, which can be
    a mixture of read-only and writable. All must share the same base_name.

    Parameters
    ----------
    prefix : str
        The PV prefix for all components of the device
    read_attrs : sequence of attribute names
        the components to include in a normal reading (i.e., in ``read()``)
    configuration_attrs : sequence of attribute names
        the components to be read less often (i.e., in
        ``read_configuration()``) and to adjust via ``configure()``
    name : str, optional
        The name of the device
    parent : instance or None
        The instance of the parent device, if applicable
    """

    SUB_ACQ_DONE = 'acq_done'  # requested acquire

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 name=None, parent=None, **kwargs):
        # Store EpicsSignal objects (only created once they are accessed)
        self._signals = {}

        self.prefix = prefix
        if self.signal_names and prefix is None:
            raise ValueError('Must specify prefix if device signals are being '
                             'used')

        if name is None:
            name = prefix

        super().__init__(name=name, parent=parent, **kwargs)

        if read_attrs is None:
            read_attrs = self.signal_names

        if configuration_attrs is None:
            configuration_attrs = []

        self.read_attrs = list(read_attrs)
        self.configuration_attrs = list(configuration_attrs)

        # Instantiate non-lazy signals
        [getattr(self, attr) for attr, cpt in self._sig_attrs.items()
         if not cpt.lazy]

    def wait_for_connection(self, all_signals=False, timeout=2.0):
        '''Wait for signals to connect

        Parameters
        ----------
        all_signals : bool, optional
            Wait for all signals to connect (including lazy ones)
        timeout : float or None
            Overall timeout
        '''
        names = [attr for attr, cpt in self._sig_attrs.items()
                 if not cpt.lazy or all_signals]

        # Instantiate first to kickoff connection process
        signals = [getattr(self, name) for name in names]

        t0 = ttime.time()
        while timeout is None or (ttime.time() - t0) < timeout:
            connected = [sig.connected for sig in signals]
            if all(connected):
                return
            ttime.sleep(min((0.05, timeout / 10.0)))

        unconnected = ', '.join(self._get_unconnected())
        raise TimeoutError('Failed to connect to all signals: {}'
                           ''.format(unconnected))

    def _get_unconnected(self):
        '''Yields all of the signal pvnames or prefixes that are unconnected

        This recurses throughout the device hierarchy, only checking signals
        that have already been instantiated.
        '''
        for attr, sig in self.get_instantiated_signals():
            if sig.connected:
                continue

            if hasattr(sig, 'pvname'):
                prefix = sig.pvname
            else:
                prefix = sig.prefix

            yield '{} ({})'.format(attr, prefix)

    def get_instantiated_signals(self, *, attr_prefix=None):
        '''Yields all of the instantiated signals in a device hierarchy

        Parameters
        ----------
        attr_prefix : string, optional
            The attribute prefix. If None, defaults to self.name

        Yields
        ------
            (fully_qualified_attribute_name, signal_instance)
        '''
        if attr_prefix is None:
            attr_prefix = self.name

        for attr, sig in self._signals.items():
            # fully qualified attribute name from top-level device
            full_attr = '{}.{}'.format(attr_prefix, attr)
            if isinstance(sig, Device):
                yield from sig.get_instantiated_signals(attr_prefix=full_attr)
            else:
                yield full_attr, sig

    @property
    def connected(self):
        return all(signal.connected for name, signal in self._signals.items())

    def __getattr__(self, name):
        '''Get a component from a fully-qualified name

        As a reminder, __getattr__ is only called if a real attribute doesn't
        already exist, or a device component has yet to be instantiated.
        '''
        if '.' not in name:
            try:
                # Initial access of signal
                cpt = self._sig_attrs[name]
                return cpt.__get__(self, None)
            except KeyError:
                raise AttributeError(name)

        attr_names = name.split('.')
        try:
            attr = getattr(self, attr_names[0])
        except AttributeError:
            raise AttributeError('{} of {}'.format(attr_names[0], name))

        if len(attr_names) > 1:
            sub_attr_names = '.'.join(attr_names[1:])
            return getattr(attr, sub_attr_names)

        return attr

    def _read_attr_list(self, attr_list, *, config=False):
        '''Get a 'read' dictionary containing attributes in attr_list'''
        values = OrderedDict()
        for attr in attr_list:
            obj = getattr(self, attr)
            if config:
                values.update(obj.read_configuration())

            values.update(obj.read())

        return values

    def read(self):
        """returns dictionary mapping names to (value, timestamp) pairs

        To control which fields are included, adjust the ``read_attrs`` list.
        """
        res = super().read()
        res.update(self._read_attr_list(self.read_attrs))
        return res

    def read_configuration(self):
        """
        returns dictionary mapping names to (value, timestamp) pairs

        To control which fields are included, adjust the
        ``configuration_attrs`` list.
        """
        return self._read_attr_list(self.configuration_attrs, config=True)

    def _describe_attr_list(self, attr_list, *, config=False):
        '''Get a 'describe' dictionary containing attributes in attr_list'''
        desc = OrderedDict()
        for attr in attr_list:
            obj = getattr(self, attr)
            if config:
                desc.update(obj.describe_configuration())

            desc.update(obj.describe())

        return desc

    def describe(self):
        '''describe the read data keys' data types and other metadata'''
        res = super().describe()
        res.update(self._describe_attr_list(self.read_attrs))
        return res

    def describe_configuration(self):
        '''describe the configuration data keys' data types/other metadata'''
        return self._describe_attr_list(self.configuration_attrs, config=True)

    @property
    def trigger_signals(self):
        names = [attr for attr, cpt in self._sig_attrs.items()
                 if cpt.trigger_value is not None]

        return [getattr(self, name) for name in names]

    def _done_acquiring(self, **kwargs):
        '''Call when acquisition has completed.'''
        self._run_subs(sub_type=self.SUB_ACQ_DONE,
                       success=True, **kwargs)

        self._reset_sub(self.SUB_ACQ_DONE)

    def trigger(self):
        """Start acquisition"""
        signals = self.trigger_signals
        if len(signals) > 1:
            raise NotImplementedError('More than one trigger signal is not '
                                      'currently supported')
        status = DeviceStatus(self)
        if not signals:
            status._finished()
            return status

        acq_signal, = signals

        self.subscribe(status._finished,
                       event_type=self.SUB_ACQ_DONE, run=False)

        def done_acquisition(**ignored_kwargs):
            # Keyword arguments are ignored here from the EpicsSignal
            # subscription, as the important part is that the put completion
            # has finished
            self._done_acquiring()

        acq_signal.put(1, wait=False, callback=done_acquisition)
        return status

    def stop(self, *, success=False):
        '''Stop the Device and all (instantiated) subdevices'''
        exc_list = []

        for attr in self._sub_devices:
            dev = getattr(self, attr)

            if not dev.connected:
                logger.debug('stop: device %s (%s) is not connected; '
                             'skipping', attr, dev)
                continue

            try:
                dev.stop(success=success)
            except ExceptionBundle as ex:
                exc_list.extend([('{}.{}'.format(attr, sub_attr), ex)
                                 for sub_attr, ex in ex.exceptions.items()])
            except Exception as ex:
                exc_list.append((attr, ex))
                logger.error('Device %s (%s) stop failed', attr, dev,
                             exc_info=ex)

        if exc_list:
            exc_info = '\n'.join('{} raised {!r}'.format(attr, ex)
                                 for attr, ex in exc_list)
            raise ExceptionBundle('{} exception(s) were raised during stop: \n'
                                  '{}'.format(len(exc_list), exc_info),
                                  exceptions=dict(exc_list))

    def get(self, **kwargs):
        '''Get the value of all components in the device

        Keyword arguments are passed onto each signal.get()
        '''
        values = {}
        for attr in self.signal_names:
            signal = getattr(self, attr)
            values[attr] = signal.get(**kwargs)

        return self._device_tuple(**values)

    def put(self, dev_t, **kwargs):
        '''Put a value to all components of the device

        Keyword arguments are passed onto each signal.put()

        Parameters
        ----------
        dev_t : DeviceTuple or tuple
            The device tuple with the value(s) to put (see get_device_tuple)
        '''
        if not isinstance(dev_t, self._device_tuple):
            try:
                dev_t = self._device_tuple(dev_t)
            except TypeError as ex:
                raise ValueError('{}\n\tDevice tuple fields: {}'
                                 ''.format(ex, self._device_tuple._fields))

        for attr in self.signal_names:
            value = getattr(dev_t, attr)
            signal = getattr(self, attr)
            signal.put(value, **kwargs)

    @classmethod
    def get_device_tuple(cls):
        '''The device tuple type associated with an Device class

        This is a tuple representing the full state of all components and
        dynamic device sub-components.
        '''
        return cls._device_tuple

    def configure(self, d):
        '''Configure the device for something during a run

        This default implementation allows the user to change any of the
        `configuration_attrs`. Subclasses might override this to perform
        additional input validation, cleanup, etc.

        Parameters
        ----------
        d : dict
            The configuration dictionary. To specify the order that
            the changes should be made, use an OrderedDict.

        Returns
        -------
        (old, new) tuple of dictionaries
        Where old and new are pre- and post-configure configuration states.
        '''
        old = self.read_configuration()
        for key, val in d.items():
            if key not in self.configuration_attrs:
                # a little extra checking for a more specific error msg
                if key not in self.signal_names:
                    raise ValueError("There is no signal named %s" % key)
                else:
                    raise ValueError("%s is not one of the "
                                     "configuration_fields, so it cannot be "
                                     "changed using configure" % key)
            set_and_wait(getattr(self, key), val)
        new = self.read_configuration()
        return old, new

    def _repr_info(self):
        yield ('prefix', self.prefix)
        yield from super()._repr_info()

        yield ('read_attrs', self.read_attrs)
        yield ('configuration_attrs', self.configuration_attrs)



"""Mixin classes that customize the filestore integration of AreaDetector
FilePlugins.

To be used like so:

    from ophyd.areadetector.detectors import PerkinElmerDetector
    from ophyd.areadetector.plugins import HDF5Plugin
    from ophyd.areadetector.trigger_mixins import SingleTrigger
    from ophyd.areadetector.filestore_mixins import (
        FileStoreIterativeWrite)

    class MyPlugin(HDF5Plugin, FileStoreIterativeWrite):
        pass

    dest = '/tmp'  # in production, use a directory on your system -- not /tmp

    class MyDetector(PerkinElmerDetector, SingleTrigger):  # for example
        file_plugin = MyPlugin(suffix='HDF1:', write_path_template=dest)

    det = MyDetector(...)
"""

import os
import logging
import warnings
import uuid

from datetime import datetime
from collections import defaultdict
from itertools import count

from ..device import GenerateDatumInterface, BlueskyInterface, Staged
from ..utils import set_and_wait

logger = logging.getLogger(__name__)


def new_uid():
    "uuid4 as a string"
    return str(uuid.uuid4())


def new_short_uid():
    "uuid4, skipping the last stanza because of AD length restrictions."
    return '-'.join(new_uid().split('-')[:-1])


def _ensure_trailing_slash(path):
    """
    'a/b/c' -> 'a/b/c/'

    EPICS adds the trailing slash itself if we do not, so in order for the
    setpoint filepath to match the readback filepath, we need to add the
    trailing slash ourselves.
    """
    return os.path.join(path, '')


class FileStoreBase(BlueskyInterface, GenerateDatumInterface):
    """Base class for FileStore mixin classes

    The `write_path_template` and `read_path_template` are formatted using
    `datetime.strftime`, which accepts the standard tokens, such as
    %Y-%m-%d.
    """
    def __init__(self, *args, fs=None, write_path_template=None,
                 read_path_template=None, **kwargs):
        if fs is None:
            warnings.warn("The device {} is not provided with a FileStore "
                          "instance. It will fall back to using the singleton "
                          "configured at import-time. This will not be "
                          "supported in future version of ophyd.".format(self))
            import filestore.api as fs
        self._fs = fs
        # TODO Can we make these args? Depends on plugin details.
        if write_path_template is None:
            raise ValueError("write_path_template is required")
        self.write_path_template = write_path_template
        self.read_path_template = read_path_template
        super().__init__(*args, **kwargs)
        self._point_counter = None
        self._locked_key_list = False
        self._datum_uids = defaultdict(list)

    @property
    def read_path_template(self):
        "Returns write_path_template if read_path_template is not set"
        if self._read_path_template is None:
            return self.write_path_template
        else:
            return self._read_path_template

    @read_path_template.setter
    def read_path_template(self, val):
        if val is not None:
            val = _ensure_trailing_slash(val)
        self._read_path_template = val

    @property
    def write_path_template(self):
        return self._write_path_template

    @write_path_template.setter
    def write_path_template(self, val):
        self._write_path_template = _ensure_trailing_slash(val)

    def stage(self):
        self._point_counter = count()
        self._locked_key_list = False
        self._datum_uids.clear()
        super().stage()

    def generate_datum(self, key, timestamp):
        "Generate a uid and cache it with its key for later insertion."
        if self._locked_key_list:
            if key not in self._datum_uids:
                raise RuntimeError("modifying after lock")
        uid = new_uid()
        reading = {'value': uid, 'timestamp': timestamp}
        # datum_uids looks like {'dark': [reading1, reading2], ...}
        self._datum_uids[key].append(reading)
        return uid

    def describe(self):
        # One object has been 'described' once, no new keys can be added
        # during this stage/unstage cycle.
        self._locked_key_list = (self._staged == Staged.yes)
        res = super().describe()
        for k in self._datum_uids:
            res[k] = self.parent.make_data_key()  # this is on DetectorBase
        return res

    def read(self):
        # One object has been 'read' once, no new keys can be added
        # during this stage/unstage cycle.
        self._locked_key_list = (self._staged == Staged.yes)
        res = super().read()
        for k, v in self._datum_uids.items():
            res[k] = v[-1]
        return res

    def unstage(self):
        self._locked_key_list = False
        self._resource = None
        return super().unstage()


class FileStorePluginBase(FileStoreBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs.update([('auto_increment', 'Yes'),
                                ('array_counter', 0),
                                ('auto_save', 'Yes'),
                                ('num_capture', 0),
                                ])
        self._fn = None

    def make_filename(self):
        '''Make a filename.

        This is a hook so that the read and write paths can either be modified
        or created on disk prior to configuring the areaDetector plugin.

        Returns
        -------
        filename : str
            The start of the filename
        read_path : str
            Path that ophyd can read from
        write_path : str
            Path that the IOC can write to
        '''
        filename = new_short_uid()
        formatter = datetime.now().strftime
        write_path = formatter(self.write_path_template)
        read_path = formatter(self.read_path_template)
        return filename, read_path, write_path

    def stage(self):
        # Make a filename.
        filename, read_path, write_path = self.make_filename()

        # Ensure we do not have an old file open.
        set_and_wait(self.capture, 0)
        # These must be set before parent is staged (specifically
        # before capture mode is turned on. They will not be reset
        # on 'unstage' anyway.
        set_and_wait(self.file_path, write_path)
        set_and_wait(self.file_name, filename)
        set_and_wait(self.file_number, 0)
        super().stage()

        # AD does this same templating in C, but we can't access it
        # so we do it redundantly here in Python.
        self._fn = self.file_template.get() % (read_path,
                                               filename,
                                               self.file_number.get() - 1)
                                               # file_number is *next* iteration
        self._fp = read_path
        if not self.file_path_exists.get():
            raise IOError("Path %s does not exist on IOC."
                          "" % self.file_path.get())


class FileStoreHDF5(FileStorePluginBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filestore_spec = 'AD_HDF5'  # spec name stored in resource doc
        self.stage_sigs.update([('file_template', '%s%s_%6.6d.h5'),
                                ('file_write_mode', 'Stream'),
                                ('capture', 1)
                                ])

    def get_frames_per_point(self):
        return self.num_capture.get()

    def stage(self):
        super().stage()
        res_kwargs = {'frame_per_point': self.get_frames_per_point()}
        logger.debug("Inserting resource with filename %s", self._fn)
        self._resource = self._fs.insert_resource(self.filestore_spec,
                                                  self._fn, res_kwargs)


class FileStoreTIFF(FileStorePluginBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filestore_spec = 'AD_TIFF'  # spec name stored in resource doc
        self.stage_sigs.update([('file_template', '%s%s_%6.6d.tiff'),
                                ('file_write_mode', 'Single'),
                                ])
        # 'Single' file_write_mode means one image : one file.
        # It does NOT mean that 'num_images' is ignored.

    def get_frames_per_point(self):
        return self.parent.cam.num_images.get()

    def stage(self):
        super().stage()
        res_kwargs = {'template': self.file_template.get(),
                      'filename': self.file_name.get(),
                      'frame_per_point': self.get_frames_per_point()}
        self._resource = self._fs.insert_resource(self.filestore_spec,
                                                  self._fp, res_kwargs)


class FileStoreTIFFSquashing(FileStorePluginBase):
    def __init__(self, *args, images_per_set_name='images_per_set',
                 number_of_sets_name="number_of_sets",
                 cam_name='cam', proc_name='proc1', **kwargs):
        super().__init__(*args, **kwargs)
        self.filestore_spec = 'AD_TIFF'  # spec name stored in resource doc
        self._ips_name = images_per_set_name
        self._num_sets_name = number_of_sets_name
        self._cam_name = cam_name
        self._proc_name = proc_name
        cam = getattr(self.parent, self._cam_name)
        proc = getattr(self.parent, self._proc_name)
        self.stage_sigs.update([('file_template', '%s%s_%6.6d.tiff'),
                                ('file_write_mode', 'Single'),
                                (proc.nd_array_port, cam.port_name.get()),
                                (proc.reset_filter, 1),
                                (proc.enable_filter, 1),
                                (proc.filter_type, 'Average'),
                                (proc.auto_reset_filter, 1),
                                (proc.filter_callbacks, 1),
                                ('nd_array_port', proc.port_name.get())
                                ])
        # 'Single' file_write_mode means one image : one file.
        # It does NOT mean that 'num_images' is ignored.

    def get_frames_per_point(self):
        return getattr(self.parent, self._num_sets_name).get()

    def stage(self):
        cam = getattr(self.parent, self._cam_name)
        proc = getattr(self.parent, self._proc_name)
        images_per_set = getattr(self.parent, self._ips_name).get()
        num_sets = getattr(self.parent, self._num_sets_name).get()

        self.stage_sigs.update([(proc.num_filter, images_per_set),
                                (cam.num_images, images_per_set * num_sets)])
        super().stage()

        res_kwargs = {'template': self.file_template.get(),
                      'filename': self.file_name.get(),
                      'frame_per_point': self.get_frames_per_point()}
        self._resource = self._fs.insert_resource(self.filestore_spec,
                                                  self._fp, res_kwargs)


class FileStoreIterativeWrite(FileStoreBase):
    "Save records to filestore as they are generated."
    def generate_datum(self, key, timestamp):
        uid = super().generate_datum(key, timestamp)
        i = next(self._point_counter)
        self._fs.insert_datum(self._resource, uid, {'point_number': i})
        return uid


class FileStoreBulkWrite(FileStoreBase):
    "Cache records as they are created and save them all at the end."
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._datum_kwargs_map = dict()  # store kwargs for each uid

    def generate_datum(self, key, timestamp):
        "Stash kwargs for each datum, to be used below by unstage."
        uid = super().generate_datum(key, timestamp)
        i = next(self._point_counter)
        self._datum_kwargs_map[uid] = {'point_number': i}
        # (don't insert, obviously)
        return uid

    def unstage(self):
        "Insert all datums at the end."
        for readings in self._datum_uids.values():
            for reading in readings:
                uid = reading['value']
                kwargs = self._datum_kwargs_map[uid]
                self._fs.insert_datum(self._resource, uid, kwargs)
        return super().unstage()


# ready-to-use combinations

class FileStoreHDF5IterativeWrite(FileStoreHDF5, FileStoreIterativeWrite):
    pass


class FileStoreHDF5BulkWrite(FileStoreHDF5, FileStoreBulkWrite):
    pass


class FileStoreTIFFIterativeWrite(FileStoreTIFF, FileStoreIterativeWrite):
    pass


class FileStoreTIFFBulkWrite(FileStoreTIFF, FileStoreBulkWrite):
    pass



# vi: ts=4 sw=4
'''
:mod:`ophyd.control.areadetector.plugins` - areaDetector plugins
======================================================

.. module:: ophyd.control.areadetector.plugins
 :synopsis:  `areaDetector`_ plugin abstractions

.. _areaDetector: http://cars.uchicago.edu/software/epics/areaDetector.html
'''


import re
import time as ttime
import logging
from collections import OrderedDict
import numpy as np

import epics

from ophyd import Component as Cpt
from .base import (ADBase, ADComponent as C, ad_group,
                   EpicsSignalWithRBV as SignalWithRBV)
from ..signal import (EpicsSignalRO, EpicsSignal, ArrayAttributeSignal)
from ..device import DynamicDeviceComponent as DDC, GenerateDatumInterface
from ..utils import enum, set_and_wait


logger = logging.getLogger(__name__)
__all__ = ['ColorConvPlugin',
           'FilePlugin',
           'HDF5Plugin',
           'ImagePlugin',
           'JPEGPlugin',
           'MagickPlugin',
           'NetCDFPlugin',
           'NexusPlugin',
           'OverlayPlugin',
           'ProcessPlugin',
           'ROIPlugin',
           'StatsPlugin',
           'TIFFPlugin',
           'TransformPlugin',
           'get_areadetector_plugin',
           'plugin_from_pvname',
           'register_plugin',
           ]


_plugin_class = {}


def register_plugin(cls):
    '''Register a plugin'''
    global _plugin_class

    _plugin_class[cls._plugin_type] = cls


class PluginBase(ADBase):
    '''AreaDetector plugin base class'''
    def __init__(self, *args, configuration_attrs=None, **kwargs):
        if configuration_attrs is None:
            configuration_attrs = self._default_configuration_attrs
        # Turn array callbacks on during staging.
        # Without this, no array data is sent to the plugins.
        super().__init__(*args, configuration_attrs=configuration_attrs,
                         **kwargs)
        # make sure it is the right type of plugin
        if (self._plugin_type is not None and
                not self.plugin_type.get().startswith(self._plugin_type)):
            raise TypeError('Trying to use {!r} class which is for {!r} '
                            'plugin type for a plugin that reports being '
                            'of type {!r} with base prefix '
                            '{!r}'.format(self.__class__.__name__,
                                          self._plugin_type,
                                          self.plugin_type.get(), self.prefix))

        self.stage_sigs['blocking_callbacks'] = 'Yes'
        if self.parent is not None and hasattr(self.parent, 'cam'):
            self.stage_sigs.update([('parent.cam.array_callbacks', 1),
                                    ])

    _default_configuration_attrs = ('port_name', 'nd_array_port', 'enable',
                                    'blocking_callbacks', 'plugin_type',
                                    'asyn_pipeline_config',
                                    'configuration_names')

    _html_docs = ['pluginDoc.html']
    _plugin_type = None
    _suffix_re = None

    array_counter = C(SignalWithRBV, 'ArrayCounter')
    array_rate = C(EpicsSignalRO, 'ArrayRate_RBV')
    asyn_io = C(EpicsSignal, 'AsynIO')

    nd_attributes_file = C(EpicsSignal, 'NDAttributesFile', string=True)
    pool_alloc_buffers = C(EpicsSignalRO, 'PoolAllocBuffers')
    pool_free_buffers = C(EpicsSignalRO, 'PoolFreeBuffers')
    pool_max_buffers = C(EpicsSignalRO, 'PoolMaxBuffers')
    pool_max_mem = C(EpicsSignalRO, 'PoolMaxMem')
    pool_used_buffers = C(EpicsSignalRO, 'PoolUsedBuffers')
    pool_used_mem = C(EpicsSignalRO, 'PoolUsedMem')
    port_name = C(EpicsSignalRO, 'PortName_RBV', string=True)

    def stage(self):
        # Ensure the plugin is enabled. We do not disable it on unstage.
        set_and_wait(self.enable, 'Enable')
        super().stage()

    @property
    def array_pixels(self):
        '''The total number of pixels, calculated from array_size'''

        array_size = self.array_size.get()
        dimensions = self.ndimensions.get()

        if dimensions == 0:
            return 0

        pixels = array_size[0]
        for dim in array_size[1:dimensions]:
            pixels *= dim

        return pixels

    def read_configuration(self):
        ret = super().read_configuration()

        ret.update(self.source_plugin.read_configuration())

        return ret

    @property
    def source_plugin(self):
        '''The PluginBase object that is the asyn source for this plugin.
        '''
        source_port = self.nd_array_port.get()
        source_plugin = self.root.get_plugin_by_asyn_port(source_port)
        return source_plugin

    def describe_configuration(self):
        ret = super().describe_configuration()

        source_plugin = self.source_plugin
        ret.update(source_plugin.describe_configuration())

        return ret

    @property
    def _asyn_pipeline(self):
        parent = self.root.get_plugin_by_asyn_port(self.nd_array_port.get())
        if hasattr(parent, '_asyn_pipeline'):
            return parent._asyn_pipeline + (self, )
        return (parent, self)

    @property
    def _asyn_pipeline_configuration_names(self):
        return [_.configuration_names.name for _ in self._asyn_pipeline]

    asyn_pipeline_config = Cpt(ArrayAttributeSignal,
                               attr='_asyn_pipeline_configuration_names')

    width = C(EpicsSignalRO, 'ArraySize0_RBV')
    height = C(EpicsSignalRO, 'ArraySize1_RBV')
    depth = C(EpicsSignalRO, 'ArraySize2_RBV')
    array_size = DDC(ad_group(EpicsSignalRO,
                              (('height', 'ArraySize1_RBV'),
                               ('width', 'ArraySize0_RBV'),
                               ('depth', 'ArraySize2_RBV'))),
                     doc='The array size')

    bayer_pattern = C(EpicsSignalRO, 'BayerPattern_RBV')
    blocking_callbacks = C(SignalWithRBV, 'BlockingCallbacks',
                           string=True)
    color_mode = C(EpicsSignalRO, 'ColorMode_RBV')
    data_type = C(EpicsSignalRO, 'DataType_RBV', string=True)

    dim0_sa = C(EpicsSignal, 'Dim0SA')
    dim1_sa = C(EpicsSignal, 'Dim1SA')
    dim2_sa = C(EpicsSignal, 'Dim2SA')
    dim_sa = DDC(ad_group(EpicsSignal,
                          (('dim0', 'Dim0SA'),
                           ('dim1', 'Dim1SA'),
                           ('dim2', 'Dim2SA'))),
                 doc='Dimension sub-arrays')

    dimensions = C(EpicsSignalRO, 'Dimensions_RBV')
    dropped_arrays = C(SignalWithRBV, 'DroppedArrays')
    enable = C(SignalWithRBV, 'EnableCallbacks', string=True)
    min_callback_time = C(SignalWithRBV, 'MinCallbackTime')
    nd_array_address = C(SignalWithRBV, 'NDArrayAddress')
    nd_array_port = C(SignalWithRBV, 'NDArrayPort')
    ndimensions = C(EpicsSignalRO, 'NDimensions_RBV')
    plugin_type = C(EpicsSignalRO, 'PluginType_RBV')

    queue_free = C(EpicsSignal, 'QueueFree')
    queue_free_low = C(EpicsSignal, 'QueueFreeLow')
    queue_size = C(EpicsSignal, 'QueueSize')
    queue_use = C(EpicsSignal, 'QueueUse')
    queue_use_high = C(EpicsSignal, 'QueueUseHIGH')
    queue_use_hihi = C(EpicsSignal, 'QueueUseHIHI')
    time_stamp = C(EpicsSignalRO, 'TimeStamp_RBV')
    unique_id = C(EpicsSignalRO, 'UniqueId_RBV')


class ImagePlugin(PluginBase):
    _default_suffix = 'image1:'
    _suffix_re = 'image\d:'
    _html_docs = ['NDPluginStdArrays.html']
    _plugin_type = 'NDPluginStdArrays'

    array_data = C(EpicsSignal, 'ArrayData')

    @property
    def image(self):
        array_size = self.array_size.get()
        if array_size == [0, 0, 0]:
            raise RuntimeError('Invalid image; ensure array_callbacks are on')

        if array_size[-1] == 0:
            array_size = array_size[:-1]

        pixel_count = self.array_pixels
        image = self.array_data.get(count=pixel_count)
        return np.array(image).reshape(array_size)


class StatsPlugin(PluginBase):
    _default_suffix = 'Stats1:'
    _suffix_re = 'Stats\d:'
    _html_docs = ['NDPluginStats.html']
    _plugin_type = 'NDPluginStats'

    _default_configuration_attrs = (PluginBase._default_configuration_attrs + (
        'centroid_threshold', 'compute_centroid', 'compute_histogram',
        'compute_profiles', 'ts_control', 'compute_statistics', 'bgd_width',
        'hist_size', 'hist_min', 'hist_max', 'ts_num_points', 'profile_size',
        'profile_cursor')
    )

    bgd_width = C(SignalWithRBV, 'BgdWidth')
    centroid_threshold = C(SignalWithRBV, 'CentroidThreshold')

    centroid = DDC(ad_group(EpicsSignalRO,
                            (('x', 'CentroidX_RBV'),
                             ('y', 'CentroidY_RBV'))),
                   doc='The centroid XY')

    compute_centroid = C(SignalWithRBV, 'ComputeCentroid', string=True)
    compute_histogram = C(SignalWithRBV, 'ComputeHistogram', string=True)
    compute_profiles = C(SignalWithRBV, 'ComputeProfiles', string=True)
    compute_statistics = C(SignalWithRBV, 'ComputeStatistics', string=True)

    cursor = DDC(ad_group(SignalWithRBV,
                          (('x', 'CursorX'),
                           ('y', 'CursorY'))),
                 doc='The cursor XY')

    hist_entropy = C(EpicsSignalRO, 'HistEntropy_RBV')
    hist_max = C(SignalWithRBV, 'HistMax')
    hist_min = C(SignalWithRBV, 'HistMin')
    hist_size = C(SignalWithRBV, 'HistSize')
    histogram = C(EpicsSignalRO, 'Histogram_RBV')

    max_size = DDC(ad_group(EpicsSignal,
                            (('x', 'MaxSizeX'),
                             ('y', 'MaxSizeY'))),
                   doc='The maximum size in XY')

    max_value = C(EpicsSignalRO, 'MaxValue_RBV')
    max_xy = DDC(ad_group(EpicsSignalRO,
                          (('x', 'MaxX_RBV'),
                           ('y', 'MaxY_RBV'))),
                 doc='Maximum in XY')

    mean_value = C(EpicsSignalRO, 'MeanValue_RBV')
    min_value = C(EpicsSignalRO, 'MinValue_RBV')

    min_xy = DDC(ad_group(EpicsSignalRO,
                          (('x', 'MinX_RBV'),
                           ('y', 'MinY_RBV'))),
                 doc='Minimum in XY')

    net = C(EpicsSignalRO, 'Net_RBV')
    profile_average = DDC(ad_group(EpicsSignalRO,
                                   (('x', 'ProfileAverageX_RBV'),
                                    ('y', 'ProfileAverageY_RBV'))),
                          doc='Profile average in XY')

    profile_centroid = DDC(ad_group(EpicsSignalRO,
                                    (('x', 'ProfileCentroidX_RBV'),
                                     ('y', 'ProfileCentroidY_RBV'))),
                           doc='Profile centroid in XY')

    profile_cursor = DDC(ad_group(EpicsSignalRO,
                                  (('x', 'ProfileCursorX_RBV'),
                                   ('y', 'ProfileCursorY_RBV'))),
                         doc='Profile cursor in XY')

    profile_size = DDC(ad_group(EpicsSignalRO,
                                (('x', 'ProfileSizeX_RBV'),
                                 ('y', 'ProfileSizeY_RBV'))),
                       doc='Profile size in XY')

    profile_threshold = DDC(ad_group(EpicsSignalRO,
                                     (('x', 'ProfileThresholdX_RBV'),
                                      ('y', 'ProfileThresholdY_RBV'))),
                            doc='Profile threshold in XY')

    set_xhopr = C(EpicsSignal, 'SetXHOPR')
    set_yhopr = C(EpicsSignal, 'SetYHOPR')
    sigma_xy = C(EpicsSignalRO, 'SigmaXY_RBV')
    sigma_x = C(EpicsSignalRO, 'SigmaX_RBV')
    sigma_y = C(EpicsSignalRO, 'SigmaY_RBV')
    sigma = C(EpicsSignalRO, 'Sigma_RBV')
    ts_acquiring = C(EpicsSignal, 'TSAcquiring')

    ts_centroid = DDC(ad_group(EpicsSignal,
                               (('x', 'TSCentroidX'),
                                ('y', 'TSCentroidY'))),
                      doc='Time series centroid in XY')

    ts_control = C(EpicsSignal, 'TSControl', string=True)
    ts_current_point = C(EpicsSignal, 'TSCurrentPoint')
    ts_max_value = C(EpicsSignal, 'TSMaxValue')

    ts_max = DDC(ad_group(EpicsSignal,
                          (('x', 'TSMaxX'),
                           ('y', 'TSMaxY'))),
                 doc='Time series maximum in XY')

    ts_mean_value = C(EpicsSignal, 'TSMeanValue')
    ts_min_value = C(EpicsSignal, 'TSMinValue')

    ts_min = DDC(ad_group(EpicsSignal,
                          (('x', 'TSMinX'),
                           ('y', 'TSMinY'))),
                 doc='Time series minimum in XY')

    ts_net = C(EpicsSignal, 'TSNet')
    ts_num_points = C(EpicsSignal, 'TSNumPoints')
    ts_read = C(EpicsSignal, 'TSRead')
    ts_sigma = C(EpicsSignal, 'TSSigma')
    ts_sigma_x = C(EpicsSignal, 'TSSigmaX')
    ts_sigma_xy = C(EpicsSignal, 'TSSigmaXY')
    ts_sigma_y = C(EpicsSignal, 'TSSigmaY')
    ts_total = C(EpicsSignal, 'TSTotal')
    total = C(EpicsSignalRO, 'Total_RBV')


class ColorConvPlugin(PluginBase):
    _default_suffix = 'CC1:'
    _suffix_re = 'CC\d:'
    _html_docs = ['NDPluginColorConvert.html']
    _plugin_type = 'NDPluginColorConvert'
    _default_configuration_attrs = (PluginBase._default_configuration_attrs +
                                    ('color_mode_out', 'false_color'))

    color_mode_out = C(SignalWithRBV, 'ColorModeOut')
    false_color = C(SignalWithRBV, 'FalseColor')


class ProcessPlugin(PluginBase):
    _default_suffix = 'Proc1:'
    _suffix_re = 'Proc\d:'
    _html_docs = ['NDPluginProcess.html']
    _plugin_type = 'NDPluginProcess'
    _default_configuration_attrs = (PluginBase._default_configuration_attrs + (
        'data_type',
        'auto_offset_scale',
        'auto_reset_filter',
        'copy_to_filter_seq',
        'data_type_out',
        'difference_seq',
        'enable_background',
        'enable_filter',
        'enable_flat_field',
        'enable_high_clip',
        'enable_low_clip',
        'enable_offset_scale',
        'fc',
        'foffset',
        'fscale',
        'filter_callbacks',
        'filter_type',
        'filter_type_seq',
        'high_clip',
        'low_clip',
        'num_filter',
        'num_filter_recip',
        'num_filtered',
        'oc',
        'o_offset',
        'o_scale',
        'offset',
        'rc',
        'roffset',
        'scale',
        'scale_flat_field',
        'valid_background',
        'valid_flat_field')
    )
    auto_offset_scale = C(EpicsSignal, 'AutoOffsetScale', string=True)
    auto_reset_filter = C(SignalWithRBV, 'AutoResetFilter', string=True)
    average_seq = C(EpicsSignal, 'AverageSeq')
    copy_to_filter_seq = C(EpicsSignal, 'CopyToFilterSeq')
    data_type_out = C(SignalWithRBV, 'DataTypeOut', string=True)
    difference_seq = C(EpicsSignal, 'DifferenceSeq')
    enable_background = C(SignalWithRBV, 'EnableBackground', string=True)
    enable_filter = C(SignalWithRBV, 'EnableFilter', string=True)
    enable_flat_field = C(SignalWithRBV, 'EnableFlatField', string=True)
    enable_high_clip = C(SignalWithRBV, 'EnableHighClip', string=True)
    enable_low_clip = C(SignalWithRBV, 'EnableLowClip', string=True)
    enable_offset_scale = C(SignalWithRBV, 'EnableOffsetScale', string=True)

    fc = DDC(ad_group(SignalWithRBV,
                      (('fc1', 'FC1'),
                       ('fc2', 'FC2'),
                       ('fc3', 'FC3'),
                       ('fc4', 'FC4'))),
             doc='Filter coefficients')

    foffset = C(SignalWithRBV, 'FOffset')
    fscale = C(SignalWithRBV, 'FScale')
    filter_callbacks = C(SignalWithRBV, 'FilterCallbacks', string=True)
    filter_type = C(EpicsSignal, 'FilterType', string=True)
    filter_type_seq = C(EpicsSignal, 'FilterTypeSeq')
    high_clip = C(SignalWithRBV, 'HighClip')
    low_clip = C(SignalWithRBV, 'LowClip')
    num_filter = C(SignalWithRBV, 'NumFilter')
    num_filter_recip = C(EpicsSignal, 'NumFilterRecip')
    num_filtered = C(EpicsSignalRO, 'NumFiltered_RBV')

    oc = DDC(ad_group(SignalWithRBV,
                      (('oc1', 'OC1'),
                       ('oc2', 'OC2'),
                       ('oc3', 'OC3'),
                       ('oc4', 'OC4'))),
             doc='Output coefficients')

    o_offset = C(SignalWithRBV, 'OOffset')
    o_scale = C(SignalWithRBV, 'OScale')
    offset = C(SignalWithRBV, 'Offset')

    rc = DDC(ad_group(SignalWithRBV,
                      (('rc1', 'RC1'),
                       ('rc2', 'RC2'))),
             doc='Filter coefficients')

    roffset = C(SignalWithRBV, 'ROffset')
    recursive_ave_diff_seq = C(EpicsSignal, 'RecursiveAveDiffSeq')
    recursive_ave_seq = C(EpicsSignal, 'RecursiveAveSeq')
    reset_filter = C(SignalWithRBV, 'ResetFilter')
    save_background = C(SignalWithRBV, 'SaveBackground')
    save_flat_field = C(SignalWithRBV, 'SaveFlatField')
    scale = C(SignalWithRBV, 'Scale')
    scale_flat_field = C(SignalWithRBV, 'ScaleFlatField')
    sum_seq = C(EpicsSignal, 'SumSeq')
    valid_background = C(EpicsSignalRO, 'ValidBackground_RBV', string=True)
    valid_flat_field = C(EpicsSignalRO, 'ValidFlatField_RBV', string=True)


class Overlay(ADBase):
    _html_docs = ['NDPluginOverlay.html']

    blue = C(SignalWithRBV, 'Blue')
    draw_mode = C(SignalWithRBV, 'DrawMode')
    green = C(SignalWithRBV, 'Green')
    max_size_x = C(EpicsSignal, 'MaxSizeX')
    max_size_y = C(EpicsSignal, 'MaxSizeY')
    overlay_portname = C(SignalWithRBV, 'Name')

    position_x = C(SignalWithRBV, 'PositionX')
    position_y = C(SignalWithRBV, 'PositionY')

    position_xlink = C(EpicsSignal, 'PositionXLink')
    position_ylink = C(EpicsSignal, 'PositionYLink')

    red = C(SignalWithRBV, 'Red')
    set_xhopr = C(EpicsSignal, 'SetXHOPR')
    set_yhopr = C(EpicsSignal, 'SetYHOPR')
    shape = C(SignalWithRBV, 'Shape')

    size_x = C(SignalWithRBV, 'SizeX')
    size_y = C(SignalWithRBV, 'SizeY')

    size_xlink = C(EpicsSignal, 'SizeXLink')
    size_ylink = C(EpicsSignal, 'SizeYLink')
    use = C(SignalWithRBV, 'Use')


class OverlayPlugin(PluginBase):
    '''Plugin which adds graphics overlays to an NDArray image

    Keyword arguments are passed to the base class, PluginBase

    Parameters
    ----------
    prefix : str
        The areaDetector plugin prefix
    '''
    _default_suffix = 'Over1:'
    _suffix_re = 'Over\d:'
    _html_docs = ['NDPluginOverlay.html']
    _plugin_type = 'NDPluginOverlay'
    _default_configuration_attrs = (PluginBase._default_configuration_attrs + (
        'overlay_1', 'overlay_2', 'overlay_3', 'overlay_4', 'overlay_5',
        'overlay_6', 'overlay_7', 'overlay_8')
    )
    max_size = DDC(ad_group(EpicsSignalRO,
                            (('x', 'MaxSizeX_RBV'),
                             ('y', 'MaxSizeY_RBV'))),
                   doc='The maximum size in XY')

    overlay_1 = C(Overlay, '1:')
    overlay_2 = C(Overlay, '2:')
    overlay_3 = C(Overlay, '3:')
    overlay_4 = C(Overlay, '4:')
    overlay_5 = C(Overlay, '5:')
    overlay_6 = C(Overlay, '6:')
    overlay_7 = C(Overlay, '7:')
    overlay_8 = C(Overlay, '8:')


class ROIPlugin(PluginBase):
    _default_suffix = 'ROI1:'
    _suffix_re = 'ROI\d:'
    _html_docs = ['NDPluginROI.html']
    _plugin_type = 'NDPluginROI'
    _default_configuration_attrs = (PluginBase._default_configuration_attrs + (
        'roi_enable', 'name_', 'bin_', 'data_type_out', 'enable_scale')
    )
    array_size = DDC(ad_group(EpicsSignalRO,
                              (('x', 'ArraySizeX_RBV'),
                               ('y', 'ArraySizeY_RBV'),
                               ('z', 'ArraySizeZ_RBV'))),
                     doc='Size of the ROI data in XYZ')

    auto_size = DDC(ad_group(SignalWithRBV,
                             (('x', 'AutoSizeX'),
                              ('y', 'AutoSizeY'),
                              ('z', 'AutoSizeZ'))),
                    doc=('Automatically set SizeXYZ to the input array size '
                         'minus MinXYZ'))

    bin_ = DDC(ad_group(SignalWithRBV,
                        (('x', 'BinX'),
                         ('y', 'BinY'),
                         ('z', 'BinZ'))),
               doc='Binning in XYZ')

    data_type_out = C(SignalWithRBV, 'DataTypeOut', string=True)
    enable_scale = C(SignalWithRBV, 'EnableScale', string=True)

    roi_enable = DDC(ad_group(SignalWithRBV,
                              (('x', 'EnableX'),
                               ('y', 'EnableY'),
                               ('z', 'EnableZ')), string=True),
                     doc=('Enable ROI calculations in the X, Y, Z dimensions. '
                          'If not enabled then the start, size, binning, and '
                          'reverse operations are disabled in the X/Y/Z '
                          'dimension, and the values from the input array '
                          'are used.'))

    max_xy = DDC(ad_group(EpicsSignal,
                          (('x', 'MaxX'),
                           ('y', 'MaxY'))),
                 doc='Maximum in XY')

    max_size = DDC(ad_group(EpicsSignalRO,
                            (('x', 'MaxSizeX_RBV'),
                             ('y', 'MaxSizeY_RBV'),
                             ('z', 'MaxSizeZ_RBV'))),
                   doc='Maximum size of the ROI in XYZ')

    min_xyz = DDC(ad_group(SignalWithRBV,
                           (('min_x', 'MinX'),
                            ('min_y', 'MinY'),
                            ('min_z', 'MinZ'))),
                  doc='Minimum size of the ROI in XYZ')

    name_ = C(SignalWithRBV, 'Name', doc='ROI name')
    reverse = DDC(ad_group(SignalWithRBV,
                           (('x', 'ReverseX'),
                            ('y', 'ReverseY'),
                            ('z', 'ReverseZ'))),
                  doc='Reverse ROI in the XYZ dimensions. (0=No, 1=Yes)')

    scale = C(SignalWithRBV, 'Scale')
    set_xhopr = C(EpicsSignal, 'SetXHOPR')
    set_yhopr = C(EpicsSignal, 'SetYHOPR')

    size = DDC(ad_group(SignalWithRBV,
                        (('x', 'SizeX'),
                         ('y', 'SizeY'),
                         ('z', 'SizeZ'))),
               doc='Size of the ROI in XYZ')


class TransformPlugin(PluginBase):
    _default_suffix = 'Trans1:'
    _suffix_re = 'Trans\d:'
    _html_docs = ['NDPluginTransform.html']
    _plugin_type = 'NDPluginTransform'

    width = C(SignalWithRBV, 'ArraySize0')
    height = C(SignalWithRBV, 'ArraySize1')
    depth = C(SignalWithRBV, 'ArraySize2')
    array_size = DDC(ad_group(SignalWithRBV,
                              (('height', 'ArraySize1'),
                               ('width', 'ArraySize0'),
                               ('depth', 'ArraySize2'))),
                     doc='Array size')

    name_ = C(EpicsSignal, 'Name')
    origin_location = C(SignalWithRBV, 'OriginLocation')
    t1_max_size = DDC(ad_group(EpicsSignal,
                               (('size0', 'T1MaxSize0'),
                                ('size1', 'T1MaxSize1'),
                                ('size2', 'T1MaxSize2'))),
                      doc='Transform 1 max size')

    t2_max_size = DDC(ad_group(EpicsSignal,
                               (('size0', 'T2MaxSize0'),
                                ('size1', 'T2MaxSize1'),
                                ('size2', 'T2MaxSize2'))),
                      doc='Transform 2 max size')

    t3_max_size = DDC(ad_group(EpicsSignal,
                               (('size0', 'T3MaxSize0'),
                                ('size1', 'T3MaxSize1'),
                                ('size2', 'T3MaxSize2'))),
                      doc='Transform 3 max size')

    t4_max_size = DDC(ad_group(EpicsSignal,
                               (('size0', 'T4MaxSize0'),
                                ('size1', 'T4MaxSize1'),
                                ('size2', 'T4MaxSize2'))),
                      doc='Transform 4 max size')

    types = DDC(ad_group(EpicsSignal,
                         (('type1', 'Type1'),
                          ('type2', 'Type2'),
                          ('type3', 'Type3'),
                          ('type4', 'Type4'))),
                doc='Transform types')


class FilePlugin(PluginBase, GenerateDatumInterface):
    _default_suffix = ''
    _html_docs = ['NDPluginFile.html']
    _plugin_type = 'NDPluginFile'
    _default_configuration_attrs = (PluginBase._default_configuration_attrs + (
        'auto_increment',
        'auto_save',
        'file_format',
        'file_name',
        'file_path',
        'file_path_exists',
        'file_template',
        'file_write_mode',
        'full_file_name',
        'num_capture'
        ))
    FileWriteMode = enum(SINGLE=0, CAPTURE=1, STREAM=2)

    auto_increment = C(SignalWithRBV, 'AutoIncrement')
    auto_save = C(SignalWithRBV, 'AutoSave')
    capture = C(SignalWithRBV, 'Capture')
    delete_driver_file = C(SignalWithRBV, 'DeleteDriverFile')
    file_format = C(SignalWithRBV, 'FileFormat')
    file_name = C(SignalWithRBV, 'FileName', string=True)
    file_number = C(SignalWithRBV, 'FileNumber')
    file_number_sync = C(EpicsSignal, 'FileNumber_Sync')
    file_number_write = C(EpicsSignal, 'FileNumber_write')
    file_path = C(SignalWithRBV, 'FilePath', string=True)
    file_path_exists = C(EpicsSignalRO, 'FilePathExists_RBV')
    file_template = C(SignalWithRBV, 'FileTemplate', string=True)
    file_write_mode = C(SignalWithRBV, 'FileWriteMode')
    full_file_name = C(EpicsSignalRO, 'FullFileName_RBV', string=True)
    num_capture = C(SignalWithRBV, 'NumCapture')
    num_captured = C(EpicsSignalRO, 'NumCaptured_RBV')
    read_file = C(SignalWithRBV, 'ReadFile')
    write_file = C(SignalWithRBV, 'WriteFile')
    write_message = C(EpicsSignal, 'WriteMessage', string=True)
    write_status = C(EpicsSignal, 'WriteStatus')


class NetCDFPlugin(FilePlugin):
    _default_suffix = 'netCDF1:'
    _suffix_re = 'netCDF\d:'
    _html_docs = ['NDFileNetCDF.html']
    _plugin_type = 'NDFileNetCDF'


class TIFFPlugin(FilePlugin):
    _default_suffix = 'TIFF1:'
    _suffix_re = 'TIFF\d:'
    _html_docs = ['NDFileTIFF.html']
    _plugin_type = 'NDFileTIFF'


class JPEGPlugin(FilePlugin):
    _default_suffix = 'JPEG1:'
    _suffix_re = 'JPEG\d:'
    _html_docs = ['NDFileJPEG.html']
    _plugin_type = 'NDFileJPEG'
    _default_configuration_attrs = (FilePlugin._default_configuration_attrs + (
        'jpeg_quality',))

    jpeg_quality = C(SignalWithRBV, 'JPEGQuality')


class NexusPlugin(FilePlugin):
    _default_suffix = 'Nexus1:'
    _suffix_re = 'Nexus\d:'
    _html_docs = ['NDFileNexus.html']
    # _plugin_type = 'NDPluginFile'  # TODO was this ever fixed?
    _plugin_type = 'NDPluginNexus'
    _default_configuration_attrs = (FilePlugin._default_configuration_attrs + (
        'template_file_name', 'template_file_path'))

    file_template_valid = C(EpicsSignal, 'FileTemplateValid')
    template_file_name = C(SignalWithRBV, 'TemplateFileName', string=True)
    template_file_path = C(SignalWithRBV, 'TemplateFilePath', string=True)


class HDF5Plugin(FilePlugin):
    _default_suffix = 'HDF1:'
    _suffix_re = 'HDF\d:'
    _html_docs = ['NDFileHDF5.html']
    _plugin_type = 'NDFileHDF5'

    _default_configuration_attrs = (FilePlugin._default_configuration_attrs + (
        'boundary_align',
        'boundary_threshold',
        'compression',
        'data_bits_offset',
        'extra_dim_name',
        'extra_dim_size',
        'io_speed',
        'num_col_chunks',
        'num_data_bits',
        'num_extra_dims',
        'num_frames_chunks',
        'num_frames_flush',
        'num_row_chunks',
        'run_time',
        'store_attr',
        'store_perform',
        'szip_num_pixels',
        'zlevel')
    )

    boundary_align = C(SignalWithRBV, 'BoundaryAlign')
    boundary_threshold = C(SignalWithRBV, 'BoundaryThreshold')
    compression = C(SignalWithRBV, 'Compression')
    data_bits_offset = C(SignalWithRBV, 'DataBitsOffset')

    extra_dim_name = DDC(ad_group(EpicsSignalRO,
                                  (('name_x', 'ExtraDimNameX_RBV'),
                                   ('name_y', 'ExtraDimNameY_RBV'),
                                   ('name_n', 'ExtraDimNameN_RBV'))),
                         doc='Extra dimension names (XYN)')
    extra_dim_size = DDC(ad_group(SignalWithRBV,
                                  (('size_x', 'ExtraDimSizeX'),
                                   ('size_y', 'ExtraDimSizeY'),
                                   ('size_n', 'ExtraDimSizeN'))),
                         doc='Extra dimension sizes (XYN)')

    io_speed = C(EpicsSignal, 'IOSpeed')
    num_col_chunks = C(SignalWithRBV, 'NumColChunks')
    num_data_bits = C(SignalWithRBV, 'NumDataBits')
    num_extra_dims = C(SignalWithRBV, 'NumExtraDims')
    num_frames_chunks = C(SignalWithRBV, 'NumFramesChunks')
    num_frames_flush = C(SignalWithRBV, 'NumFramesFlush')
    num_row_chunks = C(SignalWithRBV, 'NumRowChunks')
    run_time = C(EpicsSignal, 'RunTime')
    szip_num_pixels = C(SignalWithRBV, 'SZipNumPixels')
    store_attr = C(SignalWithRBV, 'StoreAttr')
    store_perform = C(SignalWithRBV, 'StorePerform')
    zlevel = C(SignalWithRBV, 'ZLevel')

    def warmup(self):
        """
        A convenience method for 'priming' the plugin.

        The plugin has to 'see' one acquisition before it is ready to capture.
        This sets the array size, etc.
        """
        set_and_wait(self.enable, 1)
        sigs = OrderedDict([(self.parent.cam.array_callbacks, 1),
                            (self.parent.cam.image_mode, 'Single'),
                            (self.parent.cam.trigger_mode, 'Internal'),
                            # just in case tha acquisition time is set very long...
                            (self.parent.cam.acquire_time , 1),
                            (self.parent.cam.acquire_period, 1),
                            (self.parent.cam.acquire, 1)])

        original_vals = {sig: sig.get() for sig in sigs}

        for sig, val in sigs.items():
            ttime.sleep(0.1)  # abundance of caution
            set_and_wait(sig, val)

        ttime.sleep(2)  # wait for acquisition

        for sig, val in reversed(list(original_vals.items())):
            ttime.sleep(0.1)
            set_and_wait(sig, val)


class MagickPlugin(FilePlugin):
    _default_suffix = 'Magick1:'
    _suffix_re = 'Magick\d:'
    _html_docs = ['NDFileMagick']  # sic., no html extension
    _plugin_type = 'NDFileMagick'
    _default_configuration_attrs = (FilePlugin._default_configuration_attrs + (
        'bit_depth', 'compress_type', 'quality',))

    bit_depth = C(SignalWithRBV, 'BitDepth')
    compress_type = C(SignalWithRBV, 'CompressType')
    quality = C(SignalWithRBV, 'Quality')


# register_plugin(PluginBase)
register_plugin(ImagePlugin)
register_plugin(StatsPlugin)
register_plugin(ColorConvPlugin)
register_plugin(ProcessPlugin)
register_plugin(OverlayPlugin)
register_plugin(ROIPlugin)
register_plugin(TransformPlugin)
# register_plugin(FilePlugin)
register_plugin(NetCDFPlugin)
register_plugin(TIFFPlugin)
register_plugin(JPEGPlugin)
register_plugin(NexusPlugin)
register_plugin(HDF5Plugin)
register_plugin(MagickPlugin)


def plugin_from_pvname(pv):
    '''Get the plugin class from a pvname,
    using regular expressions defined in the classes (_suffix_re).
    '''
    global _plugin_class

    for type_, cls in _plugin_class.items():
        m = re.search(cls._suffix_re, pv)
        if m:
            return cls

    return None


def get_areadetector_plugin_class(prefix, timeout=2.0):
    '''Get an areadetector plugin class by supplying its PV prefix

    Uses `plugin_from_pvname` first, but falls back on using epics channel
    access to determine the plugin type.

    Returns
    -------
    plugin : Plugin
        The plugin class

    Raises
    ------
    ValueError
        If the plugin type can't be determined
    '''
    cls = plugin_from_pvname(prefix)
    if cls is not None:
        return cls

    type_rbv = prefix + 'PluginType_RBV'
    type_ = epics.caget(type_rbv, timeout=timeout)

    if type_ is None:
        raise ValueError('Unable to determine plugin type (caget timed out)')

    # HDF5 includes version number, remove it
    type_ = type_.split(' ')[0]

    try:
        return _plugin_class[type_]
    except KeyError:
        raise ValueError('Unable to determine plugin type (PluginType={})'
                         ''.format(type_))


def get_areadetector_plugin(prefix, **kwargs):
    '''Get an instance of an areadetector plugin by supplying its PV prefix
    and any kwargs for the constructor.

    Uses `plugin_from_pvname` first, but falls back on using
    epics channel access to determine the plugin type.

    Returns
    -------
    plugin : Plugin
        The plugin instance

    Raises
    ------
    ValueError
        If the plugin type can't be determined
    '''

    cls = get_areadetector_plugin_class(prefix)
    if cls is None:
        raise ValueError('Unable to determine plugin type')

    return cls(prefix, **kwargs)





import time
import logging
import unittest
from copy import copy

import epics
from ophyd import (PVPositioner, PVPositionerPC, EpicsMotor)
from ophyd import (EpicsSignal, EpicsSignalRO)
from ophyd import (Component as C)

logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger('ophyd.pv_positioner').setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)


def tearDownModule():
    logger.debug('Cleaning up')
    logging.getLogger('ophyd.pv_positioner').setLevel(logging.INFO)
    logger.setLevel(logging.INFO)


class PVPosTest(unittest.TestCase):
    sim_pv = 'XF:31IDA-OP{Tbl-Ax:X1}Mtr'

    fake_motor = {'readback': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}-I',
                  'setpoint': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}-SP',
                  'moving': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}Sts:Moving-Sts',
                  'actuate': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}Cmd:Go-Cmd.PROC',
                  'stop': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}Cmd:Stop-Cmd.PROC',
                  }

    def test_not_subclassed(self):
        # can't instantiate it on its own
        self.assertRaises(TypeError, PVPositioner, 'prefix')
        self.assertRaises(TypeError, PVPositionerPC, 'prefix')

    def test_no_setpoint_or_readback(self):
        class MyPositioner(PVPositioner):
            pass

        self.assertRaises(ValueError, MyPositioner)

    def test_setpoint_but_no_done(self):
        class MyPositioner(PVPositioner):
            setpoint = C(EpicsSignal, '.VAL')

        self.assertRaises(ValueError, MyPositioner)

    def test_pvpos(self):
        motor_record = self.sim_pv
        mrec = EpicsMotor(motor_record, name='pvpos_mrec')
        print('mrec', mrec.describe())
        mrec.wait_for_connection()

        class MyPositioner(PVPositioner):
            '''Setpoint, readback, done, stop. No put completion'''
            setpoint = C(EpicsSignal, '.VAL')
            readback = C(EpicsSignalRO, '.RBV')
            done = C(EpicsSignalRO, '.MOVN')
            stop_signal = C(EpicsSignal, '.STOP')

            stop_value = 1
            done_value = 0

        m = MyPositioner(motor_record, name='pos_no_put_compl')
        m.wait_for_connection()

        m.read()

        mrec.move(0.1, wait=True)
        time.sleep(0.1)
        self.assertEqual(m.position, 0.1)

        m.stop()
        m.limits

        repr(m)
        str(m)

        mc = copy(m)
        self.assertEqual(mc.describe(), m.describe())

        m.read()

    def test_put_complete_setpoint_only(self):
        motor_record = self.sim_pv
        # mrec = EpicsMotor(motor_record, name='pcomplete_mrec')
        # print('mrec', mrec.describe())
        # mrec.wait_for_connection()

        logger.info('--> PV Positioner, using put completion and a DONE pv')

        class MyPositioner(PVPositionerPC):
            '''Setpoint only'''
            setpoint = C(EpicsSignal, '.VAL')

        pos = MyPositioner(motor_record, name='pc_setpoint_done')
        print(pos.describe())
        pos.wait_for_connection()

        pos.read()
        high_lim = pos.setpoint.high_limit
        try:
            pos.check_value(high_lim + 1)
        except ValueError as ex:
            logger.info('Check value for single failed, as expected (%s)', ex)
        else:
            raise ValueError('check_value should have failed')

        stat = pos.move(1, wait=False)
        logger.info('--> post-move request, moving=%s', pos.moving)

        while not stat.done:
            logger.info('--> moving... %s error=%s', stat, stat.error)
            time.sleep(0.1)

        pos.move(-1, wait=True)
        self.assertFalse(pos.moving)

    def test_put_complete_setpoint_readback_done(self):
        class MyPositioner(PVPositionerPC):
            '''Setpoint, readback, done, stop. Put completion'''
            setpoint = C(EpicsSignal, '.VAL')
            readback = C(EpicsSignalRO, '.RBV')
            done = C(EpicsSignalRO, '.MOVN')
            done_value = 0

        motor_record = self.sim_pv
        pos = MyPositioner(motor_record, name='pos_no_put_compl',
                           settle_time=0.1, timeout=25.0)
        print(pos.describe())
        pos.wait_for_connection()

        self.assertEqual(pos.settle_time, 0.1)
        self.assertEqual(pos.timeout, 25.0)
        pos.read()
        high_lim = pos.setpoint.high_limit
        try:
            pos.check_value(high_lim + 1)
        except ValueError as ex:
            logger.info('Check value for single failed, as expected (%s)', ex)
        else:
            raise ValueError('check_value should have failed')

        stat = pos.move(1, wait=False)
        self.assertEqual(stat.timeout, pos.timeout)
        logger.info('--> post-move request, moving=%s', pos.moving)

        while not stat.done:
            logger.info('--> moving... %s error=%s', stat, stat.error)
            time.sleep(0.1)

        pos.move(-1, wait=True)
        self.assertFalse(pos.moving)

    def test_put_complete_setpoint_readback(self):
        class MyPositioner(PVPositionerPC):
            '''Setpoint, readback, put completion. No done pv.'''
            setpoint = C(EpicsSignal, '.VAL')
            readback = C(EpicsSignalRO, '.RBV')

        motor_record = self.sim_pv
        pos = MyPositioner(motor_record, name='pos_put_compl')
        print(pos.describe())
        pos.wait_for_connection()

        stat = pos.move(2, wait=False)
        logger.info('--> post-move request, moving=%s', pos.moving)

        while not stat.done:
            logger.info('--> moving... %s', stat)
            time.sleep(0.1)

        pos.move(0, wait=True)
        logger.info('--> synchronous move request, moving=%s', pos.moving)

        time.sleep(0.1)
        print('read', pos.read())
        self.assertFalse(pos.moving)

    def test_pvpositioner_with_fake_motor(self):
        def callback(sub_type=None, timestamp=None, value=None, **kwargs):
            logger.info('[callback] [%s] (type=%s) value=%s', timestamp,
                        sub_type, value)

        def done_moving(value=0.0, **kwargs):
            logger.info('Done moving %s', kwargs)

        # ensure we start at 0 for this simple test
        fm = self.fake_motor
        epics.caput(fm['setpoint'], 0.05)
        time.sleep(0.5)
        epics.caput(fm['actuate'], 1)
        time.sleep(0.5)
        epics.caput(fm['setpoint'], 0)
        time.sleep(0.5)
        epics.caput(fm['actuate'], 1)
        time.sleep(0.5)

        class MyPositioner(PVPositioner):
            '''Setpoint, readback, no put completion. No done pv.'''
            setpoint = C(EpicsSignal, fm['setpoint'])
            readback = C(EpicsSignalRO, fm['readback'])
            actuate = C(EpicsSignal, fm['actuate'])
            stop_signal = C(EpicsSignal, fm['stop'])
            done = C(EpicsSignal, fm['moving'])

            actuate_value = 1
            stop_value = 1
            done_value = 1

        pos = MyPositioner('', name='pv_pos_fake_mtr')
        print('fake mtr', pos.describe())
        pos.wait_for_connection()

        pos.subscribe(callback, event_type=pos.SUB_DONE)
        pos.subscribe(callback, event_type=pos.SUB_READBACK)

        logger.info('---- test #1 ----')
        logger.info('--> move to 1')
        pos.move(1, timeout=5)
        self.assertEqual(pos.position, 1)
        logger.info('--> move to 0')
        pos.move(0, timeout=5)
        self.assertEqual(pos.position, 0)

        logger.info('---- test #2 ----')
        logger.info('--> move to 1')
        pos.move(1, wait=False)
        time.sleep(0.5)
        logger.info('--> stop')
        pos.stop()
        logger.info('--> sleep')
        time.sleep(1)
        logger.info('--> move to 0')
        pos.move(0, wait=False, moved_cb=done_moving)
        logger.info('--> post-move request, moving=%s', pos.moving)
        time.sleep(2)

        pos.read()
        repr(pos)
        str(pos)

    def test_pvpositioner_pc_with_actuate(self):
        # TODO
        self.skipTest('TODO')


from . import main
is_main = (__name__ == '__main__')
main(is_main)



import pytest
import os
import logging
import unittest
import numpy as np
import tempfile

import epics

from ophyd.utils import epics_pvs as epics_utils
from ophyd.utils import errors
from ophyd.utils import (make_dir_tree, makedirs)

from . import config

logger = logging.getLogger(__name__)


def setUpModule():
    pass


def tearDownModule():
    pass


class EpicsUtilTest(unittest.TestCase):
    def test_split(self):
        utils = epics_utils

        self.assertEquals(utils.split_record_field('record.field'),
                          ('record', 'field'))
        self.assertEquals(utils.split_record_field('record.field.invalid'),
                          ('record.field', 'invalid'))
        self.assertEquals(utils.strip_field('record.field'), 'record')
        self.assertEquals(utils.strip_field('record.field.invalid'), 'record.field')

        self.assertEquals(utils.record_field('record', 'field'), 'record.FIELD')

    def test_waveform_to_string(self):
        s = 'abcdefg'
        asc = [ord(c) for c in s]
        self.assertEquals(epics_utils.waveform_to_string(asc), s)

        asc = [ord(c) for c in s] + [0, 0, 0]
        self.assertEquals(epics_utils.waveform_to_string(asc), s)

    def test_pv_form(self):
        self.assertIn(epics_utils.get_pv_form(), ('native', 'time'))
        version = epics.__version__

        try:
            versions = ('3.2.3', '3.2.3rc1', '3.2.3-gABCD', 'unknown')
            for version in versions:
                epics.__version__ = version
                self.assertIn(epics_utils.get_pv_form(), ('native', 'time'))
        finally:
            epics.__version__ = version

    def test_records_from_db(self):
        # db_dir = os.path.join(config.epics_base, 'db')

        # if os.path.exists(db_dir):
        #     # fall back on the db file included with the tests
        db_dir = os.path.dirname(__file__)
        db_path = os.path.join(db_dir, 'scaler.db')
        records = epics_utils.records_from_db(db_path)
        self.assertIn(('bo', '$(P)$(S)_calcEnable'), records)

    def test_data_type(self):
        utils = epics_utils

        self.assertEquals(utils.data_type(1), 'integer')
        self.assertNotEqual(utils.data_type(2), 'number')
        self.assertEquals(utils.data_type(1e-3), 'number')
        self.assertEquals(utils.data_type(2.718), 'number')
        self.assertEquals(utils.data_type('foo'), 'string')
        self.assertEquals(utils.data_type(np.array([1,2,3])), 'array')

        self.assertRaises(ValueError, utils.data_type, [1,2,3])
        self.assertRaises(ValueError, utils.data_type, dict())

    def test_data_shape(self):
        utils = epics_utils

        self.assertEquals(utils.data_shape(1), list())
        self.assertEquals(utils.data_shape('foo'), list())
        self.assertEquals(utils.data_shape(np.array([1,2,3])), [3, ])
        self.assertEquals(utils.data_shape(np.array([[1, 2], [3, 4]])), [2, 2])

        self.assertRaises(ValueError, utils.data_shape, list())


def assert_OD_equal_ignore_ts(a, b):
    for (k1, v1), (k2, v2) in zip(a.items(), b.items()):
        assert (k1 == k2) and (v1['value'] == v2['value'])


def assert_file_mode(path, expected):
    assert (os.stat(path).st_mode & 0o777) == expected


def test_makedirs():
    with tempfile.TemporaryDirectory() as tempdir:
        create_dir = os.path.join(tempdir, 'a')
        makedirs(create_dir, mode=0o767, mode_base=tempdir)
        assert_file_mode(create_dir, 0o767)


def test_make_dir_tree():
    with tempfile.TemporaryDirectory() as tempdir:
        paths = make_dir_tree(2016, base_path=tempdir, mode=0o777)
        assert len(paths) == 366

        for path in paths:
            assert_file_mode(path, 0o777)

        assert os.path.join(tempdir, '2016', '03', '04') in paths
        assert os.path.join(tempdir, '2016', '02', '29') in paths


def test_valid_pvname():
    with pytest.raises(epics_utils.BadPVName):
        epics_utils.validate_pv_name('this.will.fail')


from . import main
is_main = (__name__ == '__main__')
main(is_main)



from collections import defaultdict
import time
import logging

from .status import (StatusBase, MoveStatus, DeviceStatus)

logger = logging.getLogger(__name__)


class OphydObject:
    '''The base class for all objects in Ophyd

    Handles:
    * Subscription/callback mechanism

    Parameters
    ----------
    name : str, optional
        The name of the object.
    parent : parent, optional
        The object's parent, if it exists in a hierarchy

    Attributes
    ----------
    name
    '''

    _default_sub = None

    def __init__(self, *, name=None, parent=None):
        super().__init__()

        self.name = name
        self._parent = parent

        # find magic class attributes to name 'event' types.
        # for publicly exposed event types
        self._pub_event_types = tuple(getattr(self, k) for k in dir(self) if
                                      k.startswith('SUB_'))
        # and for private event types
        self._priv_event_types = tuple(getattr(self, k) for k in dir(self) if
                                       k.startswith('_SUB_'))

        self._subs = {k: [] for k in
                      self._pub_event_types + self._priv_event_types}

        self._sub_cache = defaultdict(lambda: None)

    @property
    def connected(self):
        '''If the device is connected.

        Subclasses should override this'''
        return True

    @property
    def parent(self):
        '''The parent of the ophyd object.

        If at the top of its hierarchy, `parent` will be None
        '''
        return self._parent

    @property
    def root(self):
        "Walk parents to find ultimate ancestor (parent's parent...)."
        root = self
        while True:
            if root.parent is None:
                return root
            root = root.parent

    @property
    def report(self):
        '''A report on the object.'''
        return {}

    @property
    def event_types(self):
        '''Events that can be subscribed to via `obj.subscribe`
        '''
        return self._pub_event_types

    def _run_sub(self, cb, *args, **kwargs):
        '''Run a single subscription callback

        Parameters
        ----------
        cb
            The callback
        '''

        try:
            cb(*args, **kwargs)
        except Exception as ex:
            logger.error('Subscription %s callback exception (%s)',
                         kwargs['sub_type'],
                         self, exc_info=ex)

    def _run_cached_sub(self, sub_type, cb):
        '''Run a single subscription callback using the most recent
        cached arguments

        Parameters
        ----------
        sub_type
            The subscription type
        cb
            The callback
        '''
        cached = self._sub_cache[sub_type]
        if cached:
            args, kwargs = cached
            self._run_sub(cb, *args, **kwargs)

    def _run_subs(self, *args, sub_type, **kwargs):
        '''Run a set of subscription callbacks

        Only the kwarg ``sub_type`` is required, indicating
        the type of callback to perform. All other positional arguments
        and kwargs are passed directly to the callback function.

        No exceptions are raised when the callback functions fail.

        Parameters
        ----------
        sub_type : str
            The name of the event (sub_type) to run all of the callbacks for.
        '''
        kwargs['sub_type'] = sub_type
        kwargs.setdefault('obj', self)

        # And if a timestamp key exists, but isn't filled -- supply it with
        # a new timestamp
        if 'timestamp' in kwargs and kwargs['timestamp'] is None:
            kwargs['timestamp'] = time.time()

        # Shallow-copy the callback arguments for replaying the
        # callback at a later time (e.g., when a new subscription is made)
        self._sub_cache[sub_type] = (tuple(args), dict(kwargs))

        for cb in tuple(self._subs[sub_type]):
            self._run_sub(cb, *args, **kwargs)

    def subscribe(self, cb, event_type=None, run=True):
        '''Subscribe to events this signal group emits

        .. warning::

           If the callback raises any exceptions when run they will be
           silently ignored.

        Parameters
        ----------
        cb : callable
            A callable function (that takes kwargs) to be run when the event is
            generated.  The expected signature is ::

              def cb(*args, obj: OphydObject, sub_type: str, **kwargs) -> None:

            The exact args/kwargs passed are whatever are passed to
            ``_run_subs``
        event_type : str, optional
            The name of the event to subscribe to (if None, defaults to
            the default sub for the instance - obj._default_sub)

            This maps to the ``sub_type`` kwargs in `_run_subs`
        run : bool, optional
            Run the callback now

        See Also
        --------
        clear_sub, _run_subs

        '''
        if event_type is None:
            event_type = self._default_sub

        if event_type is None:
            raise ValueError('Subscription type not set and object {} of class'
                             ' {} has no default subscription set'
                             ''.format(self.name, self.__class__.__name__))

        try:
            self._subs[event_type].append(cb)
        except KeyError:
            raise KeyError('Unknown event type: %s' % event_type)

        if run:
            self._run_cached_sub(event_type, cb)

    def _reset_sub(self, event_type):
        '''Remove all subscriptions in an event type'''
        del self._subs[event_type][:]

    def clear_sub(self, cb, event_type=None):
        '''Remove a subscription, given the original callback function

        See also :func:`subscribe`

        Parameters
        ----------
        cb : callable
            The callback
        event_type : str, optional
            The event to unsubscribe from (if None, removes it from all event
            types)
        '''
        if event_type is None:
            for event_type, cbs in self._subs.items():
                try:
                    cbs.remove(cb)
                except ValueError:
                    pass
        else:
            self._subs[event_type].remove(cb)

    def check_value(self, value, **kwargs):
        '''Check if the value is valid for this object

        Raises
        ------
        ValueError
        '''
        pass

    def __repr__(self):
        info = self._repr_info()
        info = ', '.join('{}={!r}'.format(key, value) for key, value in info)
        return '{}({})'.format(self.__class__.__name__, info)

    def _repr_info(self):
        if self.name is not None:
            yield ('name', self.name)

        if self._parent is not None:
            yield ('parent', self.parent.name)

    def __copy__(self):
        info = dict(self._repr_info())
        return self.__class__(**info)



import logging
import time
from functools import partial
from collections import OrderedDict

from .ophydobj import OphydObject
from .status import (MoveStatus, wait as status_wait)
from .utils.epics_pvs import (data_type, data_shape)

logger = logging.getLogger(__name__)


class PositionerBase(OphydObject):
    '''The positioner base class

    Subclass from this to implement your own positioners.

    Note: Subclasses should add an additional 'wait' keyword argument on the
    move method. The MoveStatus object returned from PositionerBase can then be
    waited on after the subclass finishes the motion configuration.
    '''

    SUB_START = 'start_moving'
    SUB_DONE = 'done_moving'
    SUB_READBACK = 'readback'
    _SUB_REQ_DONE = '_req_done'  # requested move finished subscription
    _default_sub = SUB_READBACK

    def __init__(self, *, name=None, parent=None, settle_time=0.0,
                 timeout=None, **kwargs):
        super().__init__(name=name, parent=parent, **kwargs)

        self._started_moving = False
        self._moving = False
        self._position = None
        self._settle_time = settle_time
        self._timeout = timeout

    @property
    def report(self):
        rep = super().report
        rep['position'] = self.position
        return rep

    @property
    def settle_time(self):
        '''Amount of time to wait after moves to report status completion'''
        return self._settle_time

    @settle_time.setter
    def settle_time(self, settle_time):
        self._settle_time = settle_time

    @property
    def timeout(self):
        '''Amount of time to wait before to considering a motion as failed'''
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        if timeout is None:
            self._timeout = None
        else:
            self._timeout = float(timeout)

    @property
    def egu(self):
        '''The engineering units (EGU) for positions'''
        raise NotImplementedError('Subclass must implement egu')

    @property
    def limits(self):
        return (0, 0)

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def move(self, position, moved_cb=None, timeout=None):
        '''Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        '''
        if timeout is None:
            timeout = self._timeout

        self.check_value(position)

        self._run_subs(sub_type=self._SUB_REQ_DONE, success=False)
        self._reset_sub(self._SUB_REQ_DONE)

        status = MoveStatus(self, position, timeout=timeout,
                            settle_time=self._settle_time)

        if moved_cb is not None:
            status.finished_cb = partial(moved_cb, obj=self)
            # the status object will run this callback when finished

        self.subscribe(status._finished, event_type=self._SUB_REQ_DONE,
                       run=False)

        return status

    def _done_moving(self, success=True, timestamp=None, value=None, **kwargs):
        '''Call when motion has completed.  Runs SUB_DONE subscription.'''
        if success:
            self._run_subs(sub_type=self.SUB_DONE, timestamp=timestamp,
                           value=value)

        self._run_subs(sub_type=self._SUB_REQ_DONE, success=success,
                       timestamp=timestamp)
        self._reset_sub(self._SUB_REQ_DONE)

    def stop(self, *, success=False):
        '''Stops motion.

        Sub-classes must extend this method to _actually_ stop the device.

        Parameters
        ----------
        success : bool, optional
            If the move should be considered a success despite the stop.

            Defaults to False
        '''
        self._done_moving(success=success)

    @property
    def position(self):
        '''The current position of the motor in its engineering units

        Returns
        -------
        position : any
        '''
        return self._position

    def _set_position(self, value, **kwargs):
        '''Set the current internal position, run the readback subscription'''
        self._position = value

        timestamp = kwargs.pop('timestamp', time.time())
        self._run_subs(sub_type=self.SUB_READBACK, timestamp=timestamp,
                       value=value, **kwargs)

    @property
    def moving(self):
        '''Whether or not the motor is moving

        Returns
        -------
        moving : bool
        '''
        return self._moving

    def set(self, new_position, *, wait=False, moved_cb=None, timeout=None):
        """
        Bluesky-compatible API for controlling movers.

        Parameters
        ----------
        new_position : dict
            A dictionary of new positions keyed on axes name.  This is
            symmetric with read such that `mot.set(mot.read())` works as
            as expected.
        """
        return self.move(new_position, wait=wait, moved_cb=moved_cb,
                         timeout=timeout)

    def _repr_info(self):
        yield from super()._repr_info()
        yield ('settle_time', self._settle_time)
        yield ('timeout', self._timeout)


class SoftPositioner(PositionerBase):
    '''A positioner which does not communicate with any hardware

    SoftPositioner 'moves' immediately to the target position when commanded to
    do so.

    Parameters
    ----------
    limits : (low_limit, high_limit)
        Soft limits to use
    egu : str, optional
        Engineering units (EGU) for a position
    source : str, optional
        Metadata indicating the source of this positioner's position. Defaults
        to 'computed'
    '''

    def __init__(self, *, egu='', limits=None, source='computed', **kwargs):
        super().__init__(**kwargs)

        self._egu = egu
        if limits is None:
            limits = (0, 0)

        self._limits = tuple(limits)
        self.source = source

    @property
    def limits(self):
        return self._limits

    @property
    def egu(self):
        '''The engineering units (EGU) for positions'''
        return self._egu

    def _setup_move(self, position, status):
        '''Move requested to position

        This is a SoftPositioner method which allows customization of what
        happens when a motion request happens without re-implementing
        all of `move`.

        Parameters
        ----------
        position : any
            Position to move to (already verified by `check_value`)
        status : MoveStatus
            Status object created by PositionerBase.move()
        '''
        # A soft positioner immediately 'moves' to the target position when
        # requested.
        self._run_subs(sub_type=self.SUB_START, timestamp=time.time())

        self._started_moving = True
        self._moving = False

        self._set_position(position)
        self._done_moving()

    def move(self, position, wait=True, timeout=None, moved_cb=None):
        '''Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.
        wait : bool, optional
            Wait until motion has completed
        timeout : float, optional
            Maximum time to wait for a motion

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        '''
        status = super().move(position, moved_cb=moved_cb, timeout=timeout)

        self._setup_move(position, status)

        if wait:
            try:
                status_wait(status)
            except RuntimeError:
                raise RuntimeError('Motion did not complete successfully')

        return status

    def _repr_info(self):
        yield from super()._repr_info()
        yield ('egu', self._egu)
        yield ('limits', self._limits)
        yield ('source', self.source)

    def read(self):
        d = OrderedDict()
        d[self.name] = {'value': self.position,
                        'timestamp': time.time()}
        return d

    def describe(self):
        """Return the description as a dictionary

        Returns
        -------
        dict
            Dictionary of name and formatted description string
        """
        desc = OrderedDict()
        desc[self.name] = {'source': str(self.source),
                           'dtype': data_type(self.position),
                           'shape': data_shape(self.position),
                           'units': self.egu,
                           'lower_ctrl_limit': self.low_limit,
                           'upper_ctrl_limit': self.high_limit,
                           }
        return desc

    def read_configuration(self):
        return OrderedDict()

    def describe_configuration(self):
        return OrderedDict()



# vi: ts=4 sw=4
'''
:mod:`ophyd.control.pvpositioner` - Ophyd PV positioners
========================================================

.. module:: ophyd.control.pvpositioner
   :synopsis:
'''

import logging

from epics.pv import fmt_time

from .device import Device
from .positioner import PositionerBase
from .status import wait as status_wait

logger = logging.getLogger(__name__)


class PVPositioner(Device, PositionerBase):
    '''A Positioner which is controlled using multiple user-defined signals

    Keyword arguments are passed through to the base class, Positioner

    Parameters
    ----------
    prefix : str, optional
        The device prefix used for all sub-positioners. This is optional as it
        may be desirable to specify full PV names for PVPositioners.
    limits : 2-element sequence, optional
        (low_limit, high_limit)
    name : str
        The device name
    egu : str, optional
        The engineering units (EGU) for the position
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.

    Attributes
    ----------
    setpoint : Signal
        The setpoint (request) signal
    readback : Signal or None
        The readback PV (e.g., encoder position PV)
    actuate : Signal or None
        The actuation PV to set when movement is requested
    actuate_value : any, optional
        The actuation value, sent to the actuate signal when motion is
        requested
    stop_signal : Signal or None
        The stop PV to set when motion should be stopped
    stop_value : any, optional
        The value sent to stop_signal when a stop is requested
    done : Signal
        A readback value indicating whether motion is finished
    done_val : any, optional
        The value that the done pv should be when motion has completed
    put_complete : bool, optional
        If set, the specified PV should allow for asynchronous put completion
        to indicate motion has finished.  If `actuate` is specified, it will be
        used for put completion.  Otherwise, the `setpoint` will be used.  See
        the `-c` option from `caput` for more information.
    '''

    setpoint = None  # TODO: should add limits=True
    readback = None
    actuate = None
    actuate_value = 1

    stop_signal = None
    stop_value = 1

    done = None
    done_value = 1
    put_complete = False

    def __init__(self, prefix='', *, limits=None, name=None, read_attrs=None,
                 configuration_attrs=None, parent=None, egu='', **kwargs):
        super().__init__(prefix=prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs,
                         name=name, parent=parent, **kwargs)

        if self.__class__ is PVPositioner:
            raise TypeError('PVPositioner must be subclassed with the correct '
                            'signals set in the class definition.')

        self._egu = egu

        if limits is not None:
            self._limits = tuple(limits)
        else:
            self._limits = None

        if self.readback is not None:
            self.readback.subscribe(self._pos_changed)
        elif self.setpoint is not None:
            self.setpoint.subscribe(self._pos_changed)
        else:
            raise ValueError('A setpoint or a readback must be specified')

        if self.done is None and not self.put_complete:
            msg = ('PVPositioner {} is mis-configured. A "done" Signal must be'
                   ' provided or use PVPositionerPC (which uses put completion'
                   ' to determine when motion has completed).'
                   ''.format(self.name))
            raise ValueError(msg)

        if self.done is not None:
            self.done.subscribe(self._move_changed)

    @property
    def egu(self):
        '''The engineering units (EGU) for a position'''
        return self._egu

    @property
    def put_complete(self):
        return isinstance(self, PVPositionerPC)

    def check_value(self, pos):
        '''Check that the position is within the soft limits'''
        if self.limits is not None:
            low, high = self.limits
            if low != high and not (low <= pos <= high):
                raise ValueError('{} outside of user-specified limits'
                                 ''.format(pos))
        else:
            self.setpoint.check_value(pos)

    @property
    def moving(self):
        '''Whether or not the motor is moving

        If a `done` PV is specified, it will be read directly to get the motion
        status. If not, it determined from the internal state of PVPositioner.

        Returns
        -------
        bool
        '''
        if self.done is not None:
            dval = self.done.get(use_monitor=False)
            return (dval != self.done_value)
        else:
            return self._moving

    def _setup_move(self, position):
        '''Move and do not wait until motion is complete (asynchronous)'''
        logger.debug('%s.setpoint = %s', self.name, position)
        self.setpoint.put(position, wait=True)
        if self.actuate is not None:
            logger.debug('%s.actuate = %s', self.name, self.actuate_value)
            self.actuate.put(self.actuate_value, wait=False)

    def move(self, position, wait=True, timeout=None, moved_cb=None):
        '''Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        '''
        # Before moving, ensure we can stop (if a stop_signal is configured).
        if self.stop_signal is not None:
            self.stop_signal.wait_for_connection()
        status = super().move(position, timeout=timeout, moved_cb=moved_cb)

        has_done = self.done is not None
        if not has_done:
            moving_val = 1 - self.done_value
            self._move_changed(value=self.done_value)
            self._move_changed(value=moving_val)

        try:
            self._setup_move(position)
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    def _move_changed(self, timestamp=None, value=None, sub_type=None,
                      **kwargs):
        was_moving = self._moving
        self._moving = (value != self.done_value)

        started = False
        if not self._started_moving:
            started = self._started_moving = (not was_moving and self._moving)
            logger.debug('[ts=%s] %s started moving: %s', fmt_time(timestamp),
                         self.name, started)

        logger.debug('[ts=%s] %s moving: %s (value=%s)', fmt_time(timestamp),
                     self.name, self._moving, value)

        if started:
            self._run_subs(sub_type=self.SUB_START, timestamp=timestamp,
                           value=value, **kwargs)

        if not self.put_complete:
            # In the case of put completion, motion complete
            if was_moving and not self._moving:
                self._done_moving(success=True, timestamp=timestamp,
                                  value=value)

    def _pos_changed(self, timestamp=None, value=None, **kwargs):
        '''Callback from EPICS, indicating a change in position'''
        self._set_position(value)

    def stop(self, *, success=False):
        if self.stop_signal is not None:
            self.stop_signal.put(self.stop_value, wait=False)
        super().stop(success=success)

    @property
    def report(self):
        rep = super().report
        rep['pv'] = self.readback.pvname
        return rep

    @property
    def limits(self):
        if self._limits is not None:
            return tuple(self._limits)
        else:
            return self.setpoint.limits

    def _repr_info(self):
        yield from super()._repr_info()

        yield ('limits', self._limits)
        yield ('egu', self._egu)

    def _done_moving(self, **kwargs):
        has_done = self.done is not None
        if not has_done:
            self._move_changed(value=self.done_value)

        super()._done_moving(**kwargs)


class PVPositionerPC(PVPositioner):
    def __init__(self, *args, **kwargs):
        if self.__class__ is PVPositionerPC:
            raise TypeError('PVPositionerPC must be subclassed with the '
                            'correct signals set in the class definition.')

        super().__init__(*args, **kwargs)

    def _setup_move(self, position):
        '''Move and do not wait until motion is complete (asynchronous)'''
        def done_moving(**kwargs):
            logger.debug('%s async motion done', self.name)
            self._done_moving(success=True)

        if self.done is None:
            # No done signal, so we rely on put completion
            moving_val = 1 - self.done_value
            self._move_changed(value=moving_val)

        logger.debug('%s.setpoint = %s', self.name, position)

        if self.actuate is not None:
            self.setpoint.put(position, wait=True)

            logger.debug('%s.actuate = %s', self.name, self.actuate_value)
            self.actuate.put(self.actuate_value, wait=False,
                             callback=done_moving)
        else:
            self.setpoint.put(position, wait=False,
                              callback=done_moving)



from collections import OrderedDict

from . import (EpicsSignalRO, EpicsSignal, Component as Cpt,
               DynamicDeviceComponent as DDCpt, Signal)
from .areadetector import (ADComponent as ADCpt, EpicsSignalWithRBV,
                           ImagePlugin, StatsPlugin, DetectorBase,
                           SingleTrigger)


def _current_fields(attr_base, field_base, range_, **kwargs):
    defn = OrderedDict()
    for i in range_:
        attr = '{attr}{i}'.format(attr=attr_base, i=i)
        suffix = '{field}{i}'.format(field=field_base, i=i)
        defn[attr] = (EpicsSignal, suffix, kwargs)

    return defn


class QuadEM(SingleTrigger, DetectorBase):
    # This is needed because ophyd verifies that it can see all
    # of the nodes in the asyn pipeline, however these IOCs do not
    # expose their port name via a PV, but nevertheless server as the
    # root node for the plugins.
    # Leaving this port_name here for compatibility
    port_name = Cpt(Signal, value='NSLS_EM')
    model = Cpt(EpicsSignalRO, 'Model')
    firmware = Cpt(EpicsSignalRO, 'Firmware')

    acquire_mode = Cpt(EpicsSignalWithRBV, 'AcquireMode')
    acquire = Cpt(EpicsSignal, 'Acquire')

    read_format = Cpt(EpicsSignalWithRBV, 'ReadFormat')
    em_range = Cpt(EpicsSignalWithRBV, 'Range')
    ping_pong = Cpt(EpicsSignalWithRBV, 'PingPong')

    integration_time = Cpt(EpicsSignalWithRBV, 'IntegrationTime')
    num_channels = Cpt(EpicsSignalWithRBV, 'NumChannels')
    geometry = Cpt(EpicsSignalWithRBV, 'Geometry')
    resolution = Cpt(EpicsSignalWithRBV, 'Resolution')

    bias_state = Cpt(EpicsSignalWithRBV, 'BiasState')
    bias_interlock = Cpt(EpicsSignalWithRBV, 'BiasInterlock')
    bias_voltage = Cpt(EpicsSignalWithRBV, 'BiasVoltage')
    hvs_readback = Cpt(EpicsSignalRO, 'HVSReadback')
    hvv_readback = Cpt(EpicsSignalRO, 'HVVReadback')
    hvi_readback = Cpt(EpicsSignalRO, 'HVIReadback')

    values_per_read = Cpt(EpicsSignalWithRBV, 'ValuesPerRead')
    sample_time = Cpt(EpicsSignalRO, 'SampleTime_RBV')  # yay for consistency
    averaging_time = Cpt(EpicsSignalWithRBV, 'AveragingTime')
    num_average = Cpt(EpicsSignalRO, 'NumAverage_RBV')
    num_averaged = Cpt(EpicsSignalRO, 'NumAveraged_RBV')
    num_acquire = Cpt(EpicsSignalWithRBV, 'NumAcquire')
    num_acquired = Cpt(EpicsSignalRO, 'NumAcquired')
    read_data = Cpt(EpicsSignalRO, 'ReadData')
    ring_overflows = Cpt(EpicsSignalRO, 'RingOverflows')
    trigger_mode = Cpt(EpicsSignal, 'TriggerMode')
    reset = Cpt(EpicsSignal, 'Reset')

    current_names = DDCpt(_current_fields('ch', 'CurrentName', range(1, 5),
                                          string=True))
    current_offsets = DDCpt(_current_fields('ch', 'CurrentOffset',
                                            range(1, 5)))
    current_offset_calcs = DDCpt(_current_fields('ch', 'ComputeCurrentOffset',
                                                 range(1, 5)))
    current_scales = DDCpt(_current_fields('ch', 'CurrentScale', range(1, 5)))

    position_offset_x = Cpt(EpicsSignal, 'PositionOffsetX')
    position_offset_y = Cpt(EpicsSignal, 'PositionOffsetY')

    position_offset_calc_x = Cpt(EpicsSignal, 'ComputePosOffsetX')
    position_offset_calc_y = Cpt(EpicsSignal, 'ComputePosOffsetY')

    position_scale_x = Cpt(EpicsSignal, 'PositionScaleX')
    position_scale_Y = Cpt(EpicsSignal, 'PositionScaleY')

    image = ADCpt(ImagePlugin, 'image1:')
    current1 = ADCpt(StatsPlugin, 'Current1:')
    current2 = ADCpt(StatsPlugin, 'Current2:')
    current3 = ADCpt(StatsPlugin, 'Current3:')
    current4 = ADCpt(StatsPlugin, 'Current4:')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.stage_sigs.update([('acquire', 0),  # if acquiring, stop
                                ('acquire_mode', 2)  # single mode
                                ])
        self._acquisition_signal = self.acquire

        self.configuration_attrs = ['integration_time', 'averaging_time']
        self.read_attrs = ['current1.mean_value', 'current2.mean_value',
                           'current3.mean_value', 'current4.mean_value']


class NSLS_EM(QuadEM):
    port_name = Cpt(Signal, value='NSLS_EM')


class TetrAMM(QuadEM):
    port_name = Cpt(Signal, value='TetrAMM')


class APS_EM(QuadEM):
    port_name = Cpt(Signal, value='APS_EM')

